from fontTools.misc import psCharStrings
from fontTools import ttLib
from fontTools.pens.basePen import NullPen
from fontTools.misc.roundTools import otRound
from fontTools.misc.loggingTools import deprecateFunction
from fontTools.subset.util import _add_method, _uniq_sort


class _ClosureGlyphsT2Decompiler(psCharStrings.SimpleT2Decompiler):
    def __init__(self, components, localSubrs, globalSubrs):
        psCharStrings.SimpleT2Decompiler.__init__(self, localSubrs, globalSubrs)
        self.components = components

    def op_endchar(self, index):
        args = self.popall()
        if len(args) >= 4:
            from fontTools.encodings.StandardEncoding import StandardEncoding

            # endchar can do seac accent bulding; The T2 spec says it's deprecated,
            # but recent software that shall remain nameless does output it.
            adx, ady, bchar, achar = args[-4:]
            baseGlyph = StandardEncoding[bchar]
            accentGlyph = StandardEncoding[achar]
            self.components.add(baseGlyph)
            self.components.add(accentGlyph)


@_add_method(ttLib.getTableClass("CFF "))
def closure_glyphs(self, s):
    cff = self.cff
    assert len(cff) == 1
    font = cff[cff.keys()[0]]
    glyphSet = font.CharStrings

    decompose = s.glyphs
    while decompose:
        components = set()
        for g in decompose:
            if g not in glyphSet:
                continue
            gl = glyphSet[g]

            subrs = getattr(gl.private, "Subrs", [])
            decompiler = _ClosureGlyphsT2Decompiler(components, subrs, gl.globalSubrs)
            decompiler.execute(gl)
        components -= s.glyphs
        s.glyphs.update(components)
        decompose = components


def _empty_charstring(font, glyphName, isCFF2, ignoreWidth=False):
    c, fdSelectIndex = font.CharStrings.getItemAndSelector(glyphName)
    if isCFF2 or ignoreWidth:
        # CFF2 charstrings have no widths nor 'endchar' operators
        c.setProgram([] if isCFF2 else ["endchar"])
    else:
        if hasattr(font, "FDArray") and font.FDArray is not None:
            private = font.FDArray[fdSelectIndex].Private
        else:
            private = font.Private
        dfltWdX = private.defaultWidthX
        nmnlWdX = private.nominalWidthX
        pen = NullPen()
        c.draw(pen)  # this will set the charstring's width
        if c.width != dfltWdX:
            c.program = [c.width - nmnlWdX, "endchar"]
        else:
            c.program = ["endchar"]


@_add_method(ttLib.getTableClass("CFF "))
def prune_pre_subset(self, font, options):
    cff = self.cff
    # CFF table must have one font only
    cff.fontNames = cff.fontNames[:1]

    if options.notdef_glyph and not options.notdef_outline:
        isCFF2 = cff.major > 1
        for fontname in cff.keys():
            font = cff[fontname]
            _empty_charstring(font, ".notdef", isCFF2=isCFF2)

    # Clear useless Encoding
    for fontname in cff.keys():
        font = cff[fontname]
        # https://github.com/fonttools/fonttools/issues/620
        font.Encoding = "StandardEncoding"

    return True  # bool(cff.fontNames)


@_add_method(ttLib.getTableClass("CFF "))
def subset_glyphs(self, s):
    cff = self.cff
    for fontname in cff.keys():
        font = cff[fontname]
        cs = font.CharStrings

        glyphs = s.glyphs.union(s.glyphs_emptied)

        # Load all glyphs
        for g in font.charset:
            if g not in glyphs:
                continue
            c, _ = cs.getItemAndSelector(g)

        if cs.charStringsAreIndexed:
            indices = [i for i, g in enumerate(font.charset) if g in glyphs]
            csi = cs.charStringsIndex
            csi.items = [csi.items[i] for i in indices]
            del csi.file, csi.offsets
            if hasattr(font, "FDSelect"):
                sel = font.FDSelect
                sel.format = None
                sel.gidArray = [sel.gidArray[i] for i in indices]
            newCharStrings = {}
            for indicesIdx, charsetIdx in enumerate(indices):
                g = font.charset[charsetIdx]
                if g in cs.charStrings:
                    newCharStrings[g] = indicesIdx
            cs.charStrings = newCharStrings
        else:
            cs.charStrings = {g: v for g, v in cs.charStrings.items() if g in glyphs}
        font.charset = [g for g in font.charset if g in glyphs]
        font.numGlyphs = len(font.charset)

        if s.options.retain_gids:
            isCFF2 = cff.major > 1
            for g in s.glyphs_emptied:
                _empty_charstring(font, g, isCFF2=isCFF2, ignoreWidth=True)

    return True  # any(cff[fontname].numGlyphs for fontname in cff.keys())


@_add_method(ttLib.getTableClass("CFF "))
def prune_post_subset(self, ttfFont, options):
    cff = self.cff
    for fontname in cff.keys():
        font = cff[fontname]
        cs = font.CharStrings

        # Drop unused FontDictionaries
        if hasattr(font, "FDSelect"):
            sel = font.FDSelect
            indices = _uniq_sort(sel.gidArray)
            sel.gidArray = [indices.index(ss) for ss in sel.gidArray]
            arr = font.FDArray
            arr.items = [arr[i] for i in indices]
            del arr.file, arr.offsets

    # Desubroutinize if asked for
    if options.desubroutinize:
        cff.desubroutinize()

    # Drop hints if not needed
    if not options.hinting:
        self.remove_hints()
    elif not options.desubroutinize:
        self.remove_unused_subroutines()
    return True


@deprecateFunction(
    "use 'CFFFontSet.desubroutinize()' instead", category=DeprecationWarning
)
@_add_method(ttLib.getTableClass("CFF "))
def desubroutinize(self):
    self.cff.desubroutinize()


@deprecateFunction(
    "use 'CFFFontSet.remove_hints()' instead", category=DeprecationWarning
)
@_add_method(ttLib.getTableClass("CFF "))
def remove_hints(self):
    self.cff.remove_hints()


@deprecateFunction(
    "use 'CFFFontSet.remove_unused_subroutines' instead", category=DeprecationWarning
)
@_add_method(ttLib.getTableClass("CFF "))
def remove_unused_subroutines(self):
    self.cff.remove_unused_subroutines()



from .otBase import BaseTTXConverter


class table_J_S_T_F_(BaseTTXConverter):
    """Justification table

    The ``JSTF`` table contains glyph substitution and positioning
    data used to perform text justification.

    See also https://learn.microsoft.com/en-us/typography/opentype/spec/jstf
    """

    pass



from fontTools.misc.testTools import parseXML
from fontTools.misc.timeTools import timestampSinceEpoch
from fontTools.ttLib import TTFont, TTLibError
from fontTools.ttLib.tables.DefaultTable import DefaultTable
from fontTools.fontBuilder import FontBuilder
from fontTools.ttLib.tables._n_a_m_e import table__n_a_m_e
from fontTools.ttLib.tables._g_l_y_f import Glyph

from fontTools import ttx

import base64
import getopt
import logging
import os
import shutil
import subprocess
import sys
import tempfile
import unittest
from pathlib import Path

import pytest

try:
    import zopfli
except ImportError:
    zopfli = None
try:
    try:
        import brotlicffi as brotli
    except ImportError:
        import brotli
except ImportError:
    brotli = None


class TTXTest(unittest.TestCase):
    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)
        # Python 3 renamed assertRaisesRegexp to assertRaisesRegex,
        # and fires deprecation warnings if a program uses the old name.
        if not hasattr(self, "assertRaisesRegex"):
            self.assertRaisesRegex = self.assertRaisesRegexp

    def setUp(self):
        self.tempdir = None
        self.num_tempfiles = 0

    def tearDown(self):
        if self.tempdir:
            shutil.rmtree(self.tempdir)

    @staticmethod
    def getpath(testfile):
        path, _ = os.path.split(__file__)
        return os.path.join(path, "data", testfile)

    def temp_dir(self):
        if not self.tempdir:
            self.tempdir = tempfile.mkdtemp()

    def temp_font(self, font_path, file_name):
        self.temp_dir()
        temppath = os.path.join(self.tempdir, file_name)
        shutil.copy2(font_path, temppath)
        return temppath

    @staticmethod
    def read_file(file_path):
        with open(file_path, "r", encoding="utf-8") as f:
            return f.readlines()

    # -----
    # Tests
    # -----

    def test_saveXML_replace_control_characters(self):
        fb = FontBuilder(unitsPerEm=1000, isTTF=True)
        fb.setupGlyphOrder([".notdef"])
        fb.setupCharacterMap({})
        fb.setupGlyf({".notdef": Glyph()})
        fb.setupHorizontalMetrics({".notdef": (600, 0)})
        fb.setupHorizontalHeader(ascent=800, descent=-200)
        fb.setupOS2()
        fb.setupPost()
        # Setup name record with illegal control character
        nameStr = "Control\x01Char"
        replaced = nameStr.replace("\x01", "?")
        fb.setupNameTable({"familyName": nameStr})
        self.temp_dir()
        ttx_path = Path(self.tempdir) / "test.ttx"
        # Capture logs
        with self.assertLogs("fontTools.ttx", level="WARNING") as cm:
            fb.font.saveXML(str(ttx_path))
        # Test replacement
        xml_content = ttx_path.read_text(encoding="utf-8")
        self.assertIn(replaced, xml_content)
        self.assertNotIn("\x01", xml_content)
        warnings = "\n".join(cm.output)
        self.assertIn("Illegal XML character(s) found", warnings)
        self.assertIn("\\x01", warnings)

    def test_parseOptions_no_args(self):
        with self.assertRaises(getopt.GetoptError) as cm:
            ttx.parseOptions([])
        self.assertTrue("Must specify at least one input file" in str(cm.exception))

    def test_parseOptions_invalid_path(self):
        file_path = "invalid_font_path"
        with self.assertRaises(getopt.GetoptError) as cm:
            ttx.parseOptions([file_path])
        self.assertTrue('File not found: "%s"' % file_path in str(cm.exception))

    def test_parseOptions_font2ttx_1st_time(self):
        file_name = "TestOTF.otf"
        font_path = self.getpath(file_name)
        temp_path = self.temp_font(font_path, file_name)
        jobs, _ = ttx.parseOptions([temp_path])
        self.assertEqual(jobs[0][0].__name__, "ttDump")
        self.assertEqual(
            jobs[0][1:],
            (
                os.path.join(self.tempdir, file_name),
                os.path.join(self.tempdir, file_name.split(".")[0] + ".ttx"),
            ),
        )

    def test_parseOptions_font2ttx_2nd_time(self):
        file_name = "TestTTF.ttf"
        font_path = self.getpath(file_name)
        temp_path = self.temp_font(font_path, file_name)
        _, _ = ttx.parseOptions([temp_path])  # this is NOT a mistake
        jobs, _ = ttx.parseOptions([temp_path])
        self.assertEqual(jobs[0][0].__name__, "ttDump")
        self.assertEqual(
            jobs[0][1:],
            (
                os.path.join(self.tempdir, file_name),
                os.path.join(self.tempdir, file_name.split(".")[0] + "#1.ttx"),
            ),
        )

    def test_parseOptions_ttx2font_1st_time(self):
        file_name = "TestTTF.ttx"
        font_path = self.getpath(file_name)
        temp_path = self.temp_font(font_path, file_name)
        jobs, _ = ttx.parseOptions([temp_path])
        self.assertEqual(jobs[0][0].__name__, "ttCompile")
        self.assertEqual(
            jobs[0][1:],
            (
                os.path.join(self.tempdir, file_name),
                os.path.join(self.tempdir, file_name.split(".")[0] + ".ttf"),
            ),
        )

    def test_parseOptions_ttx2font_2nd_time(self):
        file_name = "TestOTF.ttx"
        font_path = self.getpath(file_name)
        temp_path = self.temp_font(font_path, file_name)
        _, _ = ttx.parseOptions([temp_path])  # this is NOT a mistake
        jobs, _ = ttx.parseOptions([temp_path])
        self.assertEqual(jobs[0][0].__name__, "ttCompile")
        self.assertEqual(
            jobs[0][1:],
            (
                os.path.join(self.tempdir, file_name),
                os.path.join(self.tempdir, file_name.split(".")[0] + "#1.otf"),
            ),
        )

    def test_parseOptions_multiple_fonts(self):
        file_names = ["TestOTF.otf", "TestTTF.ttf"]
        font_paths = [self.getpath(file_name) for file_name in file_names]
        temp_paths = [
            self.temp_font(font_path, file_name)
            for font_path, file_name in zip(font_paths, file_names)
        ]
        jobs, _ = ttx.parseOptions(temp_paths)
        for i in range(len(jobs)):
            self.assertEqual(jobs[i][0].__name__, "ttDump")
            self.assertEqual(
                jobs[i][1:],
                (
                    os.path.join(self.tempdir, file_names[i]),
                    os.path.join(self.tempdir, file_names[i].split(".")[0] + ".ttx"),
                ),
            )

    def test_parseOptions_mixed_files(self):
        operations = ["ttDump", "ttCompile"]
        extensions = [".ttx", ".ttf"]
        file_names = ["TestOTF.otf", "TestTTF.ttx"]
        font_paths = [self.getpath(file_name) for file_name in file_names]
        temp_paths = [
            self.temp_font(font_path, file_name)
            for font_path, file_name in zip(font_paths, file_names)
        ]
        jobs, _ = ttx.parseOptions(temp_paths)
        for i in range(len(jobs)):
            self.assertEqual(jobs[i][0].__name__, operations[i])
            self.assertEqual(
                jobs[i][1:],
                (
                    os.path.join(self.tempdir, file_names[i]),
                    os.path.join(
                        self.tempdir,
                        file_names[i].split(".")[0] + extensions[i],
                    ),
                ),
            )

    def test_parseOptions_splitTables(self):
        file_name = "TestTTF.ttf"
        font_path = self.getpath(file_name)
        temp_path = self.temp_font(font_path, file_name)
        args = ["-s", temp_path]

        jobs, options = ttx.parseOptions(args)

        ttx_file_path = jobs[0][2]
        temp_folder = os.path.dirname(ttx_file_path)
        self.assertTrue(options.splitTables)
        self.assertTrue(os.path.exists(ttx_file_path))

        ttx.process(jobs, options)

        # Read the TTX file but strip the first two and the last lines:
        # <?xml version="1.0" encoding="UTF-8"?>
        # <ttFont sfntVersion="\x00\x01\x00\x00" ttLibVersion="3.22">
        # ...
        # </ttFont>
        parsed_xml = parseXML(self.read_file(ttx_file_path)[2:-1])
        for item in parsed_xml:
            if not isinstance(item, tuple):
                continue
            # the tuple looks like this:
            # (u'head', {u'src': u'TestTTF._h_e_a_d.ttx'}, [])
            table_file_name = item[1].get("src")
            table_file_path = os.path.join(temp_folder, table_file_name)
            self.assertTrue(os.path.exists(table_file_path))

    def test_parseOptions_splitGlyphs(self):
        file_name = "TestTTF.ttf"
        font_path = self.getpath(file_name)
        temp_path = self.temp_font(font_path, file_name)
        args = ["-g", temp_path]

        jobs, options = ttx.parseOptions(args)

        ttx_file_path = jobs[0][2]
        temp_folder = os.path.dirname(ttx_file_path)
        self.assertTrue(options.splitGlyphs)
        # splitGlyphs also forces splitTables
        self.assertTrue(options.splitTables)
        self.assertTrue(os.path.exists(ttx_file_path))

        ttx.process(jobs, options)

        # Read the TTX file but strip the first two and the last lines:
        # <?xml version="1.0" encoding="UTF-8"?>
        # <ttFont sfntVersion="\x00\x01\x00\x00" ttLibVersion="3.22">
        # ...
        # </ttFont>
        for item in parseXML(self.read_file(ttx_file_path)[2:-1]):
            if not isinstance(item, tuple):
                continue
            # the tuple looks like this:
            # (u'head', {u'src': u'TestTTF._h_e_a_d.ttx'}, [])
            table_tag = item[0]
            table_file_name = item[1].get("src")
            table_file_path = os.path.join(temp_folder, table_file_name)
            self.assertTrue(os.path.exists(table_file_path))
            if table_tag != "glyf":
                continue
            # also strip the enclosing 'glyf' element
            for item in parseXML(self.read_file(table_file_path)[4:-3]):
                if not isinstance(item, tuple):
                    continue
                # glyphs without outline data only have 'name' attribute
                glyph_file_name = item[1].get("src")
                if glyph_file_name is not None:
                    glyph_file_path = os.path.join(temp_folder, glyph_file_name)
                    self.assertTrue(os.path.exists(glyph_file_path))

    def test_guessFileType_ttf(self):
        file_name = "TestTTF.ttf"
        font_path = self.getpath(file_name)
        self.assertEqual(ttx.guessFileType(font_path), "TTF")

    def test_guessFileType_otf(self):
        file_name = "TestOTF.otf"
        font_path = self.getpath(file_name)
        self.assertEqual(ttx.guessFileType(font_path), "OTF")

    def test_guessFileType_woff(self):
        file_name = "TestWOFF.woff"
        font_path = self.getpath(file_name)
        self.assertEqual(ttx.guessFileType(font_path), "WOFF")

    def test_guessFileType_woff2(self):
        file_name = "TestWOFF2.woff2"
        font_path = self.getpath(file_name)
        self.assertEqual(ttx.guessFileType(font_path), "WOFF2")

    def test_guessFileType_ttc(self):
        file_name = "TestTTC.ttc"
        font_path = self.getpath(file_name)
        self.assertEqual(ttx.guessFileType(font_path), "TTC")

    def test_guessFileType_dfont(self):
        file_name = "TestDFONT.dfont"
        font_path = self.getpath(file_name)
        self.assertEqual(ttx.guessFileType(font_path), "TTF")

    def test_guessFileType_ttx_ttf(self):
        file_name = "TestTTF.ttx"
        font_path = self.getpath(file_name)
        self.assertEqual(ttx.guessFileType(font_path), "TTX")

    def test_guessFileType_ttx_otf(self):
        file_name = "TestOTF.ttx"
        font_path = self.getpath(file_name)
        self.assertEqual(ttx.guessFileType(font_path), "OTX")

    def test_guessFileType_ttx_bom(self):
        file_name = "TestBOM.ttx"
        font_path = self.getpath(file_name)
        self.assertEqual(ttx.guessFileType(font_path), "TTX")

    def test_guessFileType_ttx_no_sfntVersion(self):
        file_name = "TestNoSFNT.ttx"
        font_path = self.getpath(file_name)
        self.assertEqual(ttx.guessFileType(font_path), "TTX")

    def test_guessFileType_ttx_no_xml(self):
        file_name = "TestNoXML.ttx"
        font_path = self.getpath(file_name)
        self.assertIsNone(ttx.guessFileType(font_path))

    def test_guessFileType_invalid_path(self):
        font_path = "invalid_font_path"
        self.assertIsNone(ttx.guessFileType(font_path))


# -----------------------
# ttx.Options class tests
# -----------------------


def test_options_flag_h(capsys):
    with pytest.raises(SystemExit):
        ttx.Options([("-h", None)], 1)

    out, err = capsys.readouterr()
    assert "TTX -- From OpenType To XML And Back" in out


def test_options_flag_version(capsys):
    with pytest.raises(SystemExit):
        ttx.Options([("--version", None)], 1)

    out, err = capsys.readouterr()
    version_list = out.split(".")
    assert len(version_list) >= 3
    assert version_list[0].isdigit()
    assert version_list[1].isdigit()
    assert version_list[2].strip().isdigit()


def test_options_d_goodpath(tmpdir):
    temp_dir_path = str(tmpdir)
    tto = ttx.Options([("-d", temp_dir_path)], 1)
    assert tto.outputDir == temp_dir_path


def test_options_d_badpath():
    with pytest.raises(getopt.GetoptError):
        ttx.Options([("-d", "bogusdir")], 1)


def test_options_o():
    tto = ttx.Options([("-o", "testfile.ttx")], 1)
    assert tto.outputFile == "testfile.ttx"


def test_options_f():
    tto = ttx.Options([("-f", "")], 1)
    assert tto.overWrite is True


def test_options_v():
    tto = ttx.Options([("-v", "")], 1)
    assert tto.verbose is True
    assert tto.logLevel == logging.DEBUG


def test_options_q():
    tto = ttx.Options([("-q", "")], 1)
    assert tto.quiet is True
    assert tto.logLevel == logging.WARNING


def test_options_l():
    tto = ttx.Options([("-l", "")], 1)
    assert tto.listTables is True


def test_options_t_nopadding():
    tto = ttx.Options([("-t", "CFF2")], 1)
    assert len(tto.onlyTables) == 1
    assert tto.onlyTables[0] == "CFF2"


def test_options_t_withpadding():
    tto = ttx.Options([("-t", "CFF")], 1)
    assert len(tto.onlyTables) == 1
    assert tto.onlyTables[0] == "CFF "


def test_options_s():
    tto = ttx.Options([("-s", "")], 1)
    assert tto.splitTables is True
    assert tto.splitGlyphs is False


def test_options_g():
    tto = ttx.Options([("-g", "")], 1)
    assert tto.splitGlyphs is True
    assert tto.splitTables is True


def test_options_i():
    tto = ttx.Options([("-i", "")], 1)
    assert tto.disassembleInstructions is False


def test_options_z_validoptions():
    valid_options = ("raw", "row", "bitwise", "extfile")
    for option in valid_options:
        tto = ttx.Options([("-z", option)], 1)
        assert tto.bitmapGlyphDataFormat == option


def test_options_z_invalidoption():
    with pytest.raises(getopt.GetoptError):
        ttx.Options([("-z", "bogus")], 1)


def test_options_y_validvalue():
    tto = ttx.Options([("-y", "1")], 1)
    assert tto.fontNumber == 1


def test_options_y_invalidvalue():
    with pytest.raises(ValueError):
        ttx.Options([("-y", "A")], 1)


def test_options_m():
    tto = ttx.Options([("-m", "testfont.ttf")], 1)
    assert tto.mergeFile == "testfont.ttf"


def test_options_b():
    tto = ttx.Options([("-b", "")], 1)
    assert tto.recalcBBoxes is False


def test_options_e():
    tto = ttx.Options([("-e", "")], 1)
    assert tto.ignoreDecompileErrors is False


def test_options_unicodedata():
    tto = ttx.Options([("--unicodedata", "UnicodeData.txt")], 1)
    assert tto.unicodedata == "UnicodeData.txt"


def test_options_newline_lf():
    tto = ttx.Options([("--newline", "LF")], 1)
    assert tto.newlinestr == "\n"


def test_options_newline_cr():
    tto = ttx.Options([("--newline", "CR")], 1)
    assert tto.newlinestr == "\r"


def test_options_newline_crlf():
    tto = ttx.Options([("--newline", "CRLF")], 1)
    assert tto.newlinestr == "\r\n"


def test_options_newline_invalid():
    with pytest.raises(getopt.GetoptError):
        ttx.Options([("--newline", "BOGUS")], 1)


def test_options_recalc_timestamp():
    tto = ttx.Options([("--recalc-timestamp", "")], 1)
    assert tto.recalcTimestamp is True


def test_options_no_recalc_timestamp():
    tto = ttx.Options([("--no-recalc-timestamp", "")], 1)
    assert tto.recalcTimestamp is False


def test_options_flavor():
    tto = ttx.Options([("--flavor", "woff")], 1)
    assert tto.flavor == "woff"


def test_options_with_zopfli():
    tto = ttx.Options([("--with-zopfli", ""), ("--flavor", "woff")], 1)
    assert tto.useZopfli is True


def test_options_with_zopfli_fails_without_woff_flavor():
    with pytest.raises(getopt.GetoptError):
        ttx.Options([("--with-zopfli", "")], 1)


def test_options_quiet_and_verbose_shouldfail():
    with pytest.raises(getopt.GetoptError):
        ttx.Options([("-q", ""), ("-v", "")], 1)


def test_options_mergefile_and_flavor_shouldfail():
    with pytest.raises(getopt.GetoptError):
        ttx.Options([("-m", "testfont.ttf"), ("--flavor", "woff")], 1)


def test_options_onlytables_and_skiptables_shouldfail():
    with pytest.raises(getopt.GetoptError):
        ttx.Options([("-t", "CFF"), ("-x", "CFF2")], 1)


def test_options_mergefile_and_multiplefiles_shouldfail():
    with pytest.raises(getopt.GetoptError):
        ttx.Options([("-m", "testfont.ttf")], 2)


def test_options_woff2_and_zopfli_shouldfail():
    with pytest.raises(getopt.GetoptError):
        ttx.Options([("--with-zopfli", ""), ("--flavor", "woff2")], 1)


# ----------------------------
# ttx.ttCompile function tests
# ----------------------------


def test_ttcompile_otf_compile_default(tmpdir):
    inttx = os.path.join("Tests", "ttx", "data", "TestOTF.ttx")
    # outotf = os.path.join(str(tmpdir), "TestOTF.otf")
    outotf = tmpdir.join("TestOTF.ttx")
    default_options = ttx.Options([], 1)
    ttx.ttCompile(inttx, str(outotf), default_options)
    # confirm that font was built
    assert outotf.check(file=True)
    # confirm that it is valid OTF file, can instantiate a TTFont, has expected OpenType tables
    ttf = TTFont(str(outotf))
    expected_tables = (
        "head",
        "hhea",
        "maxp",
        "OS/2",
        "name",
        "cmap",
        "post",
        "CFF ",
        "hmtx",
        "DSIG",
    )
    for table in expected_tables:
        assert table in ttf


def test_ttcompile_otf_to_woff_without_zopfli(tmpdir):
    inttx = os.path.join("Tests", "ttx", "data", "TestOTF.ttx")
    outwoff = tmpdir.join("TestOTF.woff")
    options = ttx.Options([], 1)
    options.flavor = "woff"
    ttx.ttCompile(inttx, str(outwoff), options)
    # confirm that font was built
    assert outwoff.check(file=True)
    # confirm that it is valid TTF file, can instantiate a TTFont, has expected OpenType tables
    ttf = TTFont(str(outwoff))
    expected_tables = (
        "head",
        "hhea",
        "maxp",
        "OS/2",
        "name",
        "cmap",
        "post",
        "CFF ",
        "hmtx",
        "DSIG",
    )
    for table in expected_tables:
        assert table in ttf


@pytest.mark.skipif(zopfli is None, reason="zopfli not installed")
def test_ttcompile_otf_to_woff_with_zopfli(tmpdir):
    inttx = os.path.join("Tests", "ttx", "data", "TestOTF.ttx")
    outwoff = tmpdir.join("TestOTF.woff")
    options = ttx.Options([], 1)
    options.flavor = "woff"
    options.useZopfli = True
    ttx.ttCompile(inttx, str(outwoff), options)
    # confirm that font was built
    assert outwoff.check(file=True)
    # confirm that it is valid TTF file, can instantiate a TTFont, has expected OpenType tables
    ttf = TTFont(str(outwoff))
    expected_tables = (
        "head",
        "hhea",
        "maxp",
        "OS/2",
        "name",
        "cmap",
        "post",
        "CFF ",
        "hmtx",
        "DSIG",
    )
    for table in expected_tables:
        assert table in ttf


@pytest.mark.skipif(brotli is None, reason="brotli not installed")
def test_ttcompile_otf_to_woff2(tmpdir):
    inttx = os.path.join("Tests", "ttx", "data", "TestOTF.ttx")
    outwoff2 = tmpdir.join("TestTTF.woff2")
    options = ttx.Options([], 1)
    options.flavor = "woff2"
    ttx.ttCompile(inttx, str(outwoff2), options)
    # confirm that font was built
    assert outwoff2.check(file=True)
    # confirm that it is valid TTF file, can instantiate a TTFont, has expected OpenType tables
    ttf = TTFont(str(outwoff2))
    # DSIG should not be included from original ttx as per woff2 spec (https://dev.w3.org/webfonts/WOFF2/spec/)
    assert "DSIG" not in ttf
    expected_tables = (
        "head",
        "hhea",
        "maxp",
        "OS/2",
        "name",
        "cmap",
        "post",
        "CFF ",
        "hmtx",
    )
    for table in expected_tables:
        assert table in ttf


def test_ttcompile_ttf_compile_default(tmpdir):
    inttx = os.path.join("Tests", "ttx", "data", "TestTTF.ttx")
    outttf = tmpdir.join("TestTTF.ttf")
    default_options = ttx.Options([], 1)
    ttx.ttCompile(inttx, str(outttf), default_options)
    # confirm that font was built
    assert outttf.check(file=True)
    # confirm that it is valid TTF file, can instantiate a TTFont, has expected OpenType tables
    ttf = TTFont(str(outttf))
    expected_tables = (
        "head",
        "hhea",
        "maxp",
        "OS/2",
        "name",
        "cmap",
        "hmtx",
        "fpgm",
        "prep",
        "cvt ",
        "loca",
        "glyf",
        "post",
        "gasp",
        "DSIG",
    )
    for table in expected_tables:
        assert table in ttf


def test_ttcompile_ttf_to_woff_without_zopfli(tmpdir):
    inttx = os.path.join("Tests", "ttx", "data", "TestTTF.ttx")
    outwoff = tmpdir.join("TestTTF.woff")
    options = ttx.Options([], 1)
    options.flavor = "woff"
    ttx.ttCompile(inttx, str(outwoff), options)
    # confirm that font was built
    assert outwoff.check(file=True)
    # confirm that it is valid TTF file, can instantiate a TTFont, has expected OpenType tables
    ttf = TTFont(str(outwoff))
    expected_tables = (
        "head",
        "hhea",
        "maxp",
        "OS/2",
        "name",
        "cmap",
        "hmtx",
        "fpgm",
        "prep",
        "cvt ",
        "loca",
        "glyf",
        "post",
        "gasp",
        "DSIG",
    )
    for table in expected_tables:
        assert table in ttf


@pytest.mark.skipif(zopfli is None, reason="zopfli not installed")
def test_ttcompile_ttf_to_woff_with_zopfli(tmpdir):
    inttx = os.path.join("Tests", "ttx", "data", "TestTTF.ttx")
    outwoff = tmpdir.join("TestTTF.woff")
    options = ttx.Options([], 1)
    options.flavor = "woff"
    options.useZopfli = True
    ttx.ttCompile(inttx, str(outwoff), options)
    # confirm that font was built
    assert outwoff.check(file=True)
    # confirm that it is valid TTF file, can instantiate a TTFont, has expected OpenType tables
    ttf = TTFont(str(outwoff))
    expected_tables = (
        "head",
        "hhea",
        "maxp",
        "OS/2",
        "name",
        "cmap",
        "hmtx",
        "fpgm",
        "prep",
        "cvt ",
        "loca",
        "glyf",
        "post",
        "gasp",
        "DSIG",
    )
    for table in expected_tables:
        assert table in ttf


@pytest.mark.skipif(brotli is None, reason="brotli not installed")
def test_ttcompile_ttf_to_woff2(tmpdir):
    inttx = os.path.join("Tests", "ttx", "data", "TestTTF.ttx")
    outwoff2 = tmpdir.join("TestTTF.woff2")
    options = ttx.Options([], 1)
    options.flavor = "woff2"
    ttx.ttCompile(inttx, str(outwoff2), options)
    # confirm that font was built
    assert outwoff2.check(file=True)
    # confirm that it is valid TTF file, can instantiate a TTFont, has expected OpenType tables
    ttf = TTFont(str(outwoff2))
    # DSIG should not be included from original ttx as per woff2 spec (https://dev.w3.org/webfonts/WOFF2/spec/)
    assert "DSIG" not in ttf
    expected_tables = (
        "head",
        "hhea",
        "maxp",
        "OS/2",
        "name",
        "cmap",
        "hmtx",
        "fpgm",
        "prep",
        "cvt ",
        "loca",
        "glyf",
        "post",
        "gasp",
    )
    for table in expected_tables:
        assert table in ttf


@pytest.mark.parametrize(
    "inpath, outpath1, outpath2",
    [
        ("TestTTF.ttx", "TestTTF1.ttf", "TestTTF2.ttf"),
        ("TestOTF.ttx", "TestOTF1.otf", "TestOTF2.otf"),
    ],
)
def test_ttcompile_timestamp_calcs(inpath, outpath1, outpath2, tmpdir):
    inttx = os.path.join("Tests", "ttx", "data", inpath)
    outttf1 = tmpdir.join(outpath1)
    outttf2 = tmpdir.join(outpath2)
    options = ttx.Options([], 1)
    # build with default options = do not recalculate timestamp
    ttx.ttCompile(inttx, str(outttf1), options)
    # confirm that font was built
    assert outttf1.check(file=True)
    # confirm that timestamp is same as modified time on ttx file
    mtime = os.path.getmtime(inttx)
    epochtime = timestampSinceEpoch(mtime)
    ttf = TTFont(str(outttf1))
    assert ttf["head"].modified == epochtime

    # reset options to recalculate the timestamp and compile new font
    options.recalcTimestamp = True
    ttx.ttCompile(inttx, str(outttf2), options)
    # confirm that font was built
    assert outttf2.check(file=True)
    # confirm that timestamp is more recent than modified time on ttx file
    mtime = os.path.getmtime(inttx)
    epochtime = timestampSinceEpoch(mtime)
    ttf = TTFont(str(outttf2))
    assert ttf["head"].modified > epochtime

    # --no-recalc-timestamp will keep original timestamp
    options.recalcTimestamp = False
    ttx.ttCompile(inttx, str(outttf2), options)
    assert outttf2.check(file=True)
    inttf = TTFont()
    inttf.importXML(inttx)
    assert inttf["head"].modified == TTFont(str(outttf2))["head"].modified


# -------------------------
# ttx.ttList function tests
# -------------------------


def test_ttlist_ttf(capsys, tmpdir):
    inpath = os.path.join("Tests", "ttx", "data", "TestTTF.ttf")
    fakeoutpath = tmpdir.join("TestTTF.ttx")
    options = ttx.Options([], 1)
    options.listTables = True
    ttx.ttList(inpath, str(fakeoutpath), options)
    out, err = capsys.readouterr()
    expected_tables = (
        "head",
        "hhea",
        "maxp",
        "OS/2",
        "name",
        "cmap",
        "hmtx",
        "fpgm",
        "prep",
        "cvt ",
        "loca",
        "glyf",
        "post",
        "gasp",
        "DSIG",
    )
    # confirm that expected tables are printed to stdout
    for table in expected_tables:
        assert table in out
    # test for one of the expected tag/checksum/length/offset strings
    assert "OS/2  0x67230FF8        96       376" in out


def test_ttlist_otf(capsys, tmpdir):
    inpath = os.path.join("Tests", "ttx", "data", "TestOTF.otf")
    fakeoutpath = tmpdir.join("TestOTF.ttx")
    options = ttx.Options([], 1)
    options.listTables = True
    ttx.ttList(inpath, str(fakeoutpath), options)
    out, err = capsys.readouterr()
    expected_tables = (
        "head",
        "hhea",
        "maxp",
        "OS/2",
        "name",
        "cmap",
        "post",
        "CFF ",
        "hmtx",
        "DSIG",
    )
    # confirm that expected tables are printed to stdout
    for table in expected_tables:
        assert table in out
    # test for one of the expected tag/checksum/length/offset strings
    assert "OS/2  0x67230FF8        96       272" in out


def test_ttlist_woff(capsys, tmpdir):
    inpath = os.path.join("Tests", "ttx", "data", "TestWOFF.woff")
    fakeoutpath = tmpdir.join("TestWOFF.ttx")
    options = ttx.Options([], 1)
    options.listTables = True
    options.flavor = "woff"
    ttx.ttList(inpath, str(fakeoutpath), options)
    out, err = capsys.readouterr()
    expected_tables = (
        "head",
        "hhea",
        "maxp",
        "OS/2",
        "name",
        "cmap",
        "post",
        "CFF ",
        "hmtx",
        "DSIG",
    )
    # confirm that expected tables are printed to stdout
    for table in expected_tables:
        assert table in out
    # test for one of the expected tag/checksum/length/offset strings
    assert "OS/2  0x67230FF8        84       340" in out


@pytest.mark.skipif(brotli is None, reason="brotli not installed")
def test_ttlist_woff2(capsys, tmpdir):
    inpath = os.path.join("Tests", "ttx", "data", "TestWOFF2.woff2")
    fakeoutpath = tmpdir.join("TestWOFF2.ttx")
    options = ttx.Options([], 1)
    options.listTables = True
    options.flavor = "woff2"
    ttx.ttList(inpath, str(fakeoutpath), options)
    out, err = capsys.readouterr()
    expected_tables = (
        "head",
        "hhea",
        "maxp",
        "OS/2",
        "name",
        "cmap",
        "hmtx",
        "fpgm",
        "prep",
        "cvt ",
        "loca",
        "glyf",
        "post",
        "gasp",
    )
    # confirm that expected tables are printed to stdout
    for table in expected_tables:
        assert table in out
    # test for one of the expected tag/checksum/length/offset strings
    assert "OS/2  0x67230FF8        96         0" in out


# -------------------
# main function tests
# -------------------


def test_main_default_ttf_dump_to_ttx(tmpdir):
    inpath = os.path.join("Tests", "ttx", "data", "TestTTF.ttf")
    outpath = tmpdir.join("TestTTF.ttx")
    args = ["-o", str(outpath), inpath]
    ttx.main(args)
    assert outpath.check(file=True)


def test_main_default_ttx_compile_to_ttf(tmpdir):
    inpath = os.path.join("Tests", "ttx", "data", "TestTTF.ttx")
    outpath = tmpdir.join("TestTTF.ttf")
    args = ["-o", str(outpath), inpath]
    ttx.main(args)
    assert outpath.check(file=True)


def test_main_getopterror_missing_directory():
    with pytest.raises(SystemExit):
        with pytest.raises(getopt.GetoptError):
            inpath = os.path.join("Tests", "ttx", "data", "TestTTF.ttf")
            args = ["-d", "bogusdir", inpath]
            ttx.main(args)


def test_main_keyboard_interrupt(tmpdir, monkeypatch, caplog):
    with pytest.raises(SystemExit):
        inpath = os.path.join("Tests", "ttx", "data", "TestTTF.ttx")
        outpath = tmpdir.join("TestTTF.ttf")
        args = ["-o", str(outpath), inpath]
        monkeypatch.setattr(
            ttx, "process", (lambda x, y: raise_exception(KeyboardInterrupt))
        )
        ttx.main(args)

    assert "(Cancelled.)" in caplog.text


def test_main_system_exit(tmpdir, monkeypatch):
    with pytest.raises(SystemExit):
        inpath = os.path.join("Tests", "ttx", "data", "TestTTF.ttx")
        outpath = tmpdir.join("TestTTF.ttf")
        args = ["-o", str(outpath), inpath]
        monkeypatch.setattr(ttx, "process", (lambda x, y: raise_exception(SystemExit)))
        ttx.main(args)


def test_main_ttlib_error(tmpdir, monkeypatch, caplog):
    with pytest.raises(SystemExit):
        inpath = os.path.join("Tests", "ttx", "data", "TestTTF.ttx")
        outpath = tmpdir.join("TestTTF.ttf")
        args = ["-o", str(outpath), inpath]
        monkeypatch.setattr(
            ttx,
            "process",
            (lambda x, y: raise_exception(TTLibError("Test error"))),
        )
        ttx.main(args)

    assert "Test error" in caplog.text


def test_main_base_exception(tmpdir, monkeypatch, caplog):
    with pytest.raises(SystemExit):
        inpath = os.path.join("Tests", "ttx", "data", "TestTTF.ttx")
        outpath = tmpdir.join("TestTTF.ttf")
        args = ["-o", str(outpath), inpath]
        monkeypatch.setattr(
            ttx,
            "process",
            (lambda x, y: raise_exception(Exception("Test error"))),
        )
        ttx.main(args)

    assert "Unhandled exception has occurred" in caplog.text


def test_main_ttf_dump_stdin_to_stdout(tmp_path):
    inpath = Path("Tests").joinpath("ttx", "data", "TestTTF.ttf")
    outpath = tmp_path / "TestTTF.ttx"
    args = [sys.executable, "-m", "fontTools.ttx", "-q", "-o", "-", "-"]
    with inpath.open("rb") as infile, outpath.open("w", encoding="utf-8") as outfile:
        subprocess.run(args, check=True, stdin=infile, stdout=outfile)
    assert outpath.is_file()


def test_main_ttx_compile_stdin_to_stdout(tmp_path):
    inpath = Path("Tests").joinpath("ttx", "data", "TestTTF.ttx")
    outpath = tmp_path / "TestTTF.ttf"
    args = [sys.executable, "-m", "fontTools.ttx", "-q", "-o", "-", "-"]
    with inpath.open("r", encoding="utf-8") as infile, outpath.open("wb") as outfile:
        subprocess.run(args, check=True, stdin=infile, stdout=outfile)
    assert outpath.is_file()


def test_main_gnu_style_opts_and_args_intermixed(tmpdir):
    # https://github.com/fonttools/fonttools/issues/3507
    inpath = os.path.join("Tests", "ttx", "data", "TestTTF.ttf")
    outpath = tmpdir.join("TestTTF.ttx")
    args = ["-t", "cmap", inpath, "-o", str(outpath)]
    ttx.main(args)
    assert outpath.check(file=True)


def test_roundtrip_DSIG_split_at_XML_parse_buffer_size(tmp_path):
    inpath = Path("Tests").joinpath(
        "ttx", "data", "roundtrip_DSIG_split_at_XML_parse_buffer_size.ttx"
    )
    font = TTFont()
    font.importXML(inpath)
    font["DMMY"] = DefaultTable(tag="DMMY")
    # just enough dummy bytes to hit the cut off point whereby DSIG data gets
    # split into two chunks and triggers the bug from
    # https://github.com/fonttools/fonttools/issues/2614
    font["DMMY"].data = b"\x01\x02\x03\x04" * 2438
    font.saveXML(tmp_path / "roundtrip_DSIG_split_at_XML_parse_buffer_size.ttx")

    outpath = tmp_path / "font.ttf"
    args = [
        sys.executable,
        "-m",
        "fontTools.ttx",
        "-q",
        "-o",
        str(outpath),
        str(tmp_path / "roundtrip_DSIG_split_at_XML_parse_buffer_size.ttx"),
    ]
    subprocess.run(args, check=True)

    assert outpath.is_file()
    assert TTFont(outpath)["DSIG"].signatureRecords[0].pkcs7 == base64.b64decode(
        b"0000000100000000"
    )


def test_main_ttx_compile_optimize_font_speed(tmp_path):
    inpath = Path("Tests") / "ttx" / "data" / "TestTTF.ttx"

    size_optimized = tmp_path / "TestTTF-size.ttf"
    ttx.main(["-o", str(size_optimized), str(inpath)])

    speed_optimized = tmp_path / "TestTTF-speed.ttf"
    ttx.main(["--optimize-font-speed", "-o", str(speed_optimized), str(inpath)])

    # the speed-optimized font should end up larger than the size-optimized one
    assert size_optimized.stat().st_size < speed_optimized.stat().st_size


# ---------------------------
# support functions for tests
# ---------------------------


def raise_exception(exception):
    raise exception



from fontTools.misc.xmlWriter import XMLWriter
from fontTools.ttLib import TTFont
from fontTools.feaLib.lookupDebugInfo import LOOKUP_DEBUG_ENV_VAR
from fontTools import mtiLib
import difflib
from io import StringIO
import os
import sys
import pytest


@pytest.fixture(autouse=True)
def set_lookup_debug_env_var(monkeypatch):
    monkeypatch.setenv(LOOKUP_DEBUG_ENV_VAR, "1")


class MtiTest:
    GLYPH_ORDER = [
        ".notdef",
        "a",
        "b",
        "pakannada",
        "phakannada",
        "vakannada",
        "pevowelkannada",
        "phevowelkannada",
        "vevowelkannada",
        "uvowelsignkannada",
        "uuvowelsignkannada",
        "uvowelsignaltkannada",
        "uuvowelsignaltkannada",
        "uuvowelsignsinh",
        "uvowelsignsinh",
        "rakarsinh",
        "zero",
        "one",
        "two",
        "three",
        "four",
        "five",
        "six",
        "seven",
        "eight",
        "nine",
        "slash",
        "fraction",
        "A",
        "B",
        "C",
        "fi",
        "fl",
        "breve",
        "acute",
        "uniFB01",
        "ffi",
        "grave",
        "commaacent",
        "dotbelow",
        "dotabove",
        "cedilla",
        "commaaccent",
        "Acircumflex",
        "V",
        "T",
        "acircumflex",
        "Aacute",
        "Agrave",
        "O",
        "Oacute",
        "Ograve",
        "Ocircumflex",
        "aacute",
        "agrave",
        "aimatrabindigurmukhi",
        "aimatragurmukhi",
        "aimatratippigurmukhi",
        "aumatrabindigurmukhi",
        "aumatragurmukhi",
        "bindigurmukhi",
        "eematrabindigurmukhi",
        "eematragurmukhi",
        "eematratippigurmukhi",
        "oomatrabindigurmukhi",
        "oomatragurmukhi",
        "oomatratippigurmukhi",
        "lagurmukhi",
        "lanuktagurmukhi",
        "nagurmukhi",
        "nanuktagurmukhi",
        "ngagurmukhi",
        "nganuktagurmukhi",
        "nnagurmukhi",
        "nnanuktagurmukhi",
        "tthagurmukhi",
        "tthanuktagurmukhi",
        "bsuperior",
        "isuperior",
        "vsuperior",
        "wsuperior",
        "periodsuperior",
        "osuperior",
        "tsuperior",
        "dollarsuperior",
        "fsuperior",
        "gsuperior",
        "zsuperior",
        "dsuperior",
        "psuperior",
        "hsuperior",
        "oesuperior",
        "aesuperior",
        "centsuperior",
        "esuperior",
        "lsuperior",
        "qsuperior",
        "csuperior",
        "asuperior",
        "commasuperior",
        "xsuperior",
        "egravesuperior",
        "usuperior",
        "rsuperior",
        "nsuperior",
        "ssuperior",
        "msuperior",
        "jsuperior",
        "ysuperior",
        "ksuperior",
        "guilsinglright",
        "guilsinglleft",
        "uniF737",
        "uniE11C",
        "uniE11D",
        "uniE11A",
        "uni2077",
        "uni2087",
        "uniE11B",
        "uniE119",
        "uniE0DD",
        "uniE0DE",
        "uniF736",
        "uniE121",
        "uniE122",
        "uniE11F",
        "uni2076",
        "uni2086",
        "uniE120",
        "uniE11E",
        "uniE0DB",
        "uniE0DC",
        "uniF733",
        "uniE12B",
        "uniE12C",
        "uniE129",
        "uni00B3",
        "uni2083",
        "uniE12A",
        "uniE128",
        "uniF732",
        "uniE133",
        "uniE134",
        "uniE131",
        "uni00B2",
        "uni2082",
        "uniE132",
        "uniE130",
        "uniE0F9",
        "uniF734",
        "uniE0D4",
        "uniE0D5",
        "uniE0D2",
        "uni2074",
        "uni2084",
        "uniE0D3",
        "uniE0D1",
        "uniF730",
        "uniE13D",
        "uniE13E",
        "uniE13A",
        "uni2070",
        "uni2080",
        "uniE13B",
        "uniE139",
        "uniE13C",
        "uniF739",
        "uniE0EC",
        "uniE0ED",
        "uniE0EA",
        "uni2079",
        "uni2089",
        "uniE0EB",
        "uniE0E9",
        "uniF735",
        "uniE0CD",
        "uniE0CE",
        "uniE0CB",
        "uni2075",
        "uni2085",
        "uniE0CC",
        "uniE0CA",
        "uniF731",
        "uniE0F3",
        "uniE0F4",
        "uniE0F1",
        "uni00B9",
        "uni2081",
        "uniE0F2",
        "uniE0F0",
        "uniE0F8",
        "uniF738",
        "uniE0C0",
        "uniE0C1",
        "uniE0BE",
        "uni2078",
        "uni2088",
        "uniE0BF",
        "uniE0BD",
        "I",
        "Ismall",
        "t",
        "i",
        "f",
        "IJ",
        "J",
        "IJsmall",
        "Jsmall",
        "tt",
        "ij",
        "j",
        "ffb",
        "ffh",
        "h",
        "ffk",
        "k",
        "ffl",
        "l",
        "fft",
        "fb",
        "ff",
        "fh",
        "fj",
        "fk",
        "ft",
        "janyevoweltelugu",
        "kassevoweltelugu",
        "jaivoweltelugu",
        "nyasubscripttelugu",
        "kaivoweltelugu",
        "ssasubscripttelugu",
        "bayi1",
        "jeemi1",
        "kafi1",
        "ghafi1",
        "laami1",
        "kafm1",
        "ghafm1",
        "laamm1",
        "rayf2",
        "reyf2",
        "yayf2",
        "zayf2",
        "fayi1",
        "ayehf2",
        "hamzayeharabf2",
        "hamzayehf2",
        "yehf2",
        "ray",
        "rey",
        "zay",
        "yay",
        "dal",
        "del",
        "zal",
        "rayf1",
        "reyf1",
        "yayf1",
        "zayf1",
        "ayehf1",
        "hamzayeharabf1",
        "hamzayehf1",
        "yehf1",
        "dal1",
        "del1",
        "zal1",
        "onehalf",
        "onehalf.alt",
        "onequarter",
        "onequarter.alt",
        "threequarters",
        "threequarters.alt",
        "AlefSuperiorNS",
        "DammaNS",
        "DammaRflxNS",
        "DammatanNS",
        "Fatha2dotsNS",
        "FathaNS",
        "FathatanNS",
        "FourDotsAboveNS",
        "HamzaAboveNS",
        "MaddaNS",
        "OneDotAbove2NS",
        "OneDotAboveNS",
        "ShaddaAlefNS",
        "ShaddaDammaNS",
        "ShaddaDammatanNS",
        "ShaddaFathatanNS",
        "ShaddaKasraNS",
        "ShaddaKasratanNS",
        "ShaddaNS",
        "SharetKafNS",
        "SukunNS",
        "ThreeDotsDownAboveNS",
        "ThreeDotsUpAboveNS",
        "TwoDotsAboveNS",
        "TwoDotsVerticalAboveNS",
        "UltapeshNS",
        "WaslaNS",
        "AinIni.12m_MeemFin.02",
        "AinIni_YehBarreeFin",
        "AinMed_YehBarreeFin",
        "BehxIni_MeemFin",
        "BehxIni_NoonGhunnaFin",
        "BehxIni_RehFin",
        "BehxIni_RehFin.b",
        "BehxMed_MeemFin.py",
        "BehxMed_NoonGhunnaFin",
        "BehxMed_NoonGhunnaFin.cup",
        "BehxMed_RehFin",
        "BehxMed_RehFin.cup",
        "BehxMed_YehxFin",
        "FehxMed_YehBarreeFin",
        "HahIni_YehBarreeFin",
        "KafIni_YehBarreeFin",
        "KafMed.12_YehxFin.01",
        "KafMed_MeemFin",
        "KafMed_YehBarreeFin",
        "LamAlefFin",
        "LamAlefFin.cup",
        "LamAlefFin.cut",
        "LamAlefFin.short",
        "LamAlefSep",
        "LamIni_MeemFin",
        "LamIni_YehBarreeFin",
        "LamMed_MeemFin",
        "LamMed_MeemFin.b",
        "LamMed_YehxFin",
        "LamMed_YehxFin.cup",
        "TahIni_YehBarreeFin",
        "null",
        "CR",
        "space",
        "exclam",
        "quotedbl",
        "numbersign",
    ]

    # Feature files in data/*.txt; output gets compared to data/*.ttx.
    TESTS = {
        None: ("mti/cmap",),
        "cmap": ("mti/cmap",),
        "GSUB": (
            "featurename-backward",
            "featurename-forward",
            "lookupnames-backward",
            "lookupnames-forward",
            "mixed-toplevels",
            "mti/scripttable",
            "mti/chainedclass",
            "mti/chainedcoverage",
            "mti/chained-glyph",
            "mti/gsubalternate",
            "mti/gsubligature",
            "mti/gsubmultiple",
            "mti/gsubreversechanined",
            "mti/gsubsingle",
        ),
        "GPOS": (
            "mti/scripttable",
            "mti/chained-glyph",
            "mti/gposcursive",
            "mti/gposkernset",
            "mti/gposmarktobase",
            "mti/gpospairclass",
            "mti/gpospairglyph",
            "mti/gpossingle",
            "mti/mark-to-ligature",
        ),
        "GDEF": (
            "mti/gdefattach",
            "mti/gdefclasses",
            "mti/gdefligcaret",
            "mti/gdefmarkattach",
            "mti/gdefmarkfilter",
        ),
    }
    # TODO:
    # https://github.com/Monotype/OpenType_Table_Source/issues/12
    #
    #        'mti/featuretable'
    #        'mti/contextclass'
    #        'mti/contextcoverage'
    #        'mti/context-glyph'

    @staticmethod
    def getpath(testfile):
        path, _ = os.path.split(__file__)
        return os.path.join(path, "data", testfile)

    def expect_ttx(self, expected_ttx, actual_ttx, fromfile=None, tofile=None):
        expected = [l + "\n" for l in expected_ttx.split("\n")]
        actual = [l + "\n" for l in actual_ttx.split("\n")]
        if actual != expected:
            sys.stderr.write("\n")
            for line in difflib.unified_diff(
                expected, actual, fromfile=fromfile, tofile=tofile
            ):
                sys.stderr.write(line)
            pytest.fail("TTX output is different from expected")

    @classmethod
    def create_font(celf):
        font = TTFont()
        font.setGlyphOrder(celf.GLYPH_ORDER)
        return font

    def check_mti_file(self, name, tableTag=None):
        xml_expected_path = self.getpath(
            "%s.ttx" % name + ("." + tableTag if tableTag is not None else "")
        )
        with open(xml_expected_path, "rt", encoding="utf-8") as xml_expected_file:
            xml_expected = xml_expected_file.read()

        font = self.create_font()

        with open(self.getpath("%s.txt" % name), "rt", encoding="utf-8") as f:
            table = mtiLib.build(f, font, tableTag=tableTag)

        if tableTag is not None:
            assert tableTag == table.tableTag
        tableTag = table.tableTag

        # Make sure it compiles.
        blob = table.compile(font)

        # Make sure it decompiles.
        decompiled = table.__class__()
        decompiled.decompile(blob, font)

        # XML from built object.
        writer = XMLWriter(StringIO())
        writer.begintag(tableTag)
        writer.newline()
        table.toXML(writer, font)
        writer.endtag(tableTag)
        writer.newline()
        xml_built = writer.file.getvalue()

        # XML from decompiled object.
        writer = XMLWriter(StringIO())
        writer.begintag(tableTag)
        writer.newline()
        decompiled.toXML(writer, font)
        writer.endtag(tableTag)
        writer.newline()
        xml_binary = writer.file.getvalue()

        self.expect_ttx(xml_binary, xml_built, fromfile="decompiled", tofile="built")
        self.expect_ttx(
            xml_expected, xml_built, fromfile=xml_expected_path, tofile="built"
        )

        from fontTools.misc import xmlReader

        f = StringIO()
        f.write(xml_expected)
        f.seek(0)
        font2 = TTFont()
        font2.setGlyphOrder(font.getGlyphOrder())
        reader = xmlReader.XMLReader(f, font2)
        reader.read(rootless=True)

        # XML from object read from XML.
        writer = XMLWriter(StringIO())
        writer.begintag(tableTag)
        writer.newline()
        font2[tableTag].toXML(writer, font)
        writer.endtag(tableTag)
        writer.newline()
        xml_fromxml = writer.file.getvalue()

        self.expect_ttx(
            xml_expected, xml_fromxml, fromfile=xml_expected_path, tofile="fromxml"
        )


def generate_mti_file_test(name, tableTag=None):
    return lambda self: self.check_mti_file(
        os.path.join(*name.split("/")), tableTag=tableTag
    )


for tableTag, tests in MtiTest.TESTS.items():
    for name in tests:
        setattr(
            MtiTest,
            "test_MtiFile_%s%s" % (name, "_" + tableTag if tableTag else ""),
            generate_mti_file_test(name, tableTag=tableTag),
        )


if __name__ == "__main__":
    if len(sys.argv) > 1:
        from fontTools.mtiLib import main

        font = MtiTest.create_font()
        sys.exit(main(sys.argv[1:], font))
    sys.exit(pytest.main(sys.argv))



#!/usr/bin/env python3
""" Convert SVG paths to UFO glyphs. """


__requires__ = ["fontTools"]

from types import SimpleNamespace
from fontTools.svgLib import SVGPath

from fontTools.pens.pointPen import SegmentToPointPen
from fontTools.ufoLib.glifLib import writeGlyphToString


__all__ = ["svg2glif"]


def svg2glif(svg, name, width=0, height=0, unicodes=None, transform=None, version=2):
    """Convert an SVG outline to a UFO glyph with given 'name', advance
    'width' and 'height' (int), and 'unicodes' (list of int).
    Return the resulting string in GLIF format (default: version 2).
    If 'transform' is provided, apply a transformation matrix before the
    conversion (must be tuple of 6 floats, or a FontTools Transform object).
    """
    glyph = SimpleNamespace(width=width, height=height, unicodes=unicodes)
    outline = SVGPath.fromstring(svg, transform=transform)

    # writeGlyphToString takes a callable (usually a glyph's drawPoints
    # method) that accepts a PointPen, however SVGPath currently only has
    # a draw method that accepts a segment pen. We need to wrap the call
    # with a converter pen.
    def drawPoints(pointPen):
        pen = SegmentToPointPen(pointPen)
        outline.draw(pen)

    return writeGlyphToString(
        name, glyphObject=glyph, drawPointsFunc=drawPoints, formatVersion=version
    )


def parse_args(args):
    import argparse

    def split(arg):
        return arg.replace(",", " ").split()

    def unicode_hex_list(arg):
        try:
            return [int(unihex, 16) for unihex in split(arg)]
        except ValueError:
            msg = "Invalid unicode hexadecimal value: %r" % arg
            raise argparse.ArgumentTypeError(msg)

    def transform_list(arg):
        try:
            return [float(n) for n in split(arg)]
        except ValueError:
            msg = "Invalid transformation matrix: %r" % arg
            raise argparse.ArgumentTypeError(msg)

    parser = argparse.ArgumentParser(
        description="Convert SVG outlines to UFO glyphs (.glif)"
    )
    parser.add_argument(
        "infile",
        metavar="INPUT.svg",
        help="Input SVG file containing " '<path> elements with "d" attributes.',
    )
    parser.add_argument(
        "outfile",
        metavar="OUTPUT.glif",
        help="Output GLIF file (default: " "print to stdout)",
        nargs="?",
    )
    parser.add_argument(
        "-n",
        "--name",
        help="The glyph name (default: input SVG file "
        "basename, without the .svg extension)",
    )
    parser.add_argument(
        "-w",
        "--width",
        help="The glyph advance width (default: 0)",
        type=int,
        default=0,
    )
    parser.add_argument(
        "-H",
        "--height",
        help="The glyph vertical advance (optional if " '"width" is defined)',
        type=int,
        default=0,
    )
    parser.add_argument(
        "-u",
        "--unicodes",
        help="List of Unicode code points as hexadecimal "
        'numbers (e.g. -u "0041 0042")',
        type=unicode_hex_list,
    )
    parser.add_argument(
        "-t",
        "--transform",
        help="Transformation matrix as a list of six "
        'float values (e.g. -t "0.1 0 0 -0.1 -50 200")',
        type=transform_list,
    )
    parser.add_argument(
        "-f",
        "--format",
        help="UFO GLIF format version (default: 2)",
        type=int,
        choices=(1, 2),
        default=2,
    )

    return parser.parse_args(args)


def main(args=None):
    from io import open

    options = parse_args(args)

    svg_file = options.infile

    if options.name:
        name = options.name
    else:
        import os

        name = os.path.splitext(os.path.basename(svg_file))[0]

    with open(svg_file, "r", encoding="utf-8-sig") as f:
        svg = f.read()

    glif = svg2glif(
        svg,
        name,
        width=options.width,
        height=options.height,
        unicodes=options.unicodes,
        transform=options.transform,
        version=options.format,
    )

    if options.outfile is None:
        print(glif)
    else:
        with open(options.outfile, "w", encoding="utf-8") as f:
            f.write(glif)


if __name__ == "__main__":
    import sys

    sys.exit(main())



"""Enum-related utilities, including backports for older Python versions."""

from __future__ import annotations

from enum import Enum


__all__ = ["StrEnum"]

# StrEnum is only available in Python 3.11+
try:
    from enum import StrEnum
except ImportError:

    class StrEnum(str, Enum):
        """
        Minimal backport of Python 3.11's StrEnum for older versions.

        An Enum where all members are also strings.
        """

        def __str__(self) -> str:
            return self.value



from fontTools.pens.boundsPen import BoundsPen, ControlBoundsPen
import unittest


def draw_(pen):
    pen.moveTo((0, 0))
    pen.lineTo((0, 100))
    pen.qCurveTo((50, 75), (60, 50), (50, 25), (0, 0))
    pen.curveTo((-50, 25), (-60, 50), (-50, 75), (0, 100))
    pen.closePath()


def bounds_(pen):
    return " ".join(["%.0f" % c for c in pen.bounds])


class BoundsPenTest(unittest.TestCase):
    def test_draw(self):
        pen = BoundsPen(None)
        draw_(pen)
        self.assertEqual("-55 0 58 100", bounds_(pen))

    def test_empty(self):
        pen = BoundsPen(None)
        self.assertEqual(None, pen.bounds)

    def test_curve(self):
        pen = BoundsPen(None)
        pen.moveTo((0, 0))
        pen.curveTo((20, 10), (90, 40), (0, 0))
        self.assertEqual("0 0 45 20", bounds_(pen))

    def test_quadraticCurve(self):
        pen = BoundsPen(None)
        pen.moveTo((0, 0))
        pen.qCurveTo((6, 6), (10, 0))
        self.assertEqual("0 0 10 3", bounds_(pen))


class ControlBoundsPenTest(unittest.TestCase):
    def test_draw(self):
        pen = ControlBoundsPen(None)
        draw_(pen)
        self.assertEqual("-55 0 60 100", bounds_(pen))

    def test_empty(self):
        pen = ControlBoundsPen(None)
        self.assertEqual(None, pen.bounds)

    def test_curve(self):
        pen = ControlBoundsPen(None)
        pen.moveTo((0, 0))
        pen.curveTo((20, 10), (90, 40), (0, 0))
        self.assertEqual("0 0 90 40", bounds_(pen))

    def test_quadraticCurve(self):
        pen = ControlBoundsPen(None)
        pen.moveTo((0, 0))
        pen.qCurveTo((6, 6), (10, 0))
        self.assertEqual("0 0 10 6", bounds_(pen))

    def test_singlePoint(self):
        pen = ControlBoundsPen(None)
        pen.moveTo((-5, 10))
        self.assertEqual("-5 10 -5 10", bounds_(pen))

    def test_ignoreSinglePoint(self):
        pen = ControlBoundsPen(None, ignoreSinglePoints=True)
        pen.moveTo((0, 10))
        self.assertEqual(None, pen.bounds)


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



from fontTools.voltLib import ast
from fontTools.voltLib.error import VoltLibError
from fontTools.voltLib.parser import Parser
from io import StringIO
import unittest


class ParserTest(unittest.TestCase):
    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)
        # Python 3 renamed assertRaisesRegexp to assertRaisesRegex,
        # and fires deprecation warnings if a program uses the old name.
        if not hasattr(self, "assertRaisesRegex"):
            self.assertRaisesRegex = self.assertRaisesRegexp

    def assertSubEqual(self, sub, glyph_ref, replacement_ref):
        glyphs = [[g.glyph for g in v] for v in sub.mapping.keys()]
        replacement = [[g.glyph for g in v] for v in sub.mapping.values()]

        self.assertEqual(glyphs, glyph_ref)
        self.assertEqual(replacement, replacement_ref)

    def test_def_glyph_base(self):
        [def_glyph] = self.parse(
            'DEF_GLYPH ".notdef" ID 0 TYPE BASE END_GLYPH'
        ).statements
        self.assertEqual(
            (
                def_glyph.name,
                def_glyph.id,
                def_glyph.unicode,
                def_glyph.type,
                def_glyph.components,
            ),
            (".notdef", 0, None, "BASE", None),
        )

    def test_def_glyph_base_with_unicode(self):
        [def_glyph] = self.parse(
            'DEF_GLYPH "space" ID 3 UNICODE 32 TYPE BASE END_GLYPH'
        ).statements
        self.assertEqual(
            (
                def_glyph.name,
                def_glyph.id,
                def_glyph.unicode,
                def_glyph.type,
                def_glyph.components,
            ),
            ("space", 3, [0x0020], "BASE", None),
        )

    def test_def_glyph_base_with_unicodevalues(self):
        [def_glyph] = self.parse_(
            'DEF_GLYPH "CR" ID 2 UNICODEVALUES "U+0009" ' "TYPE BASE END_GLYPH"
        ).statements
        self.assertEqual(
            (
                def_glyph.name,
                def_glyph.id,
                def_glyph.unicode,
                def_glyph.type,
                def_glyph.components,
            ),
            ("CR", 2, [0x0009], "BASE", None),
        )

    def test_def_glyph_base_with_mult_unicodevalues(self):
        [def_glyph] = self.parse(
            'DEF_GLYPH "CR" ID 2 UNICODEVALUES "U+0009,U+000D" ' "TYPE BASE END_GLYPH"
        ).statements
        self.assertEqual(
            (
                def_glyph.name,
                def_glyph.id,
                def_glyph.unicode,
                def_glyph.type,
                def_glyph.components,
            ),
            ("CR", 2, [0x0009, 0x000D], "BASE", None),
        )

    def test_def_glyph_base_with_empty_unicodevalues(self):
        [def_glyph] = self.parse_(
            'DEF_GLYPH "i.locl" ID 269 UNICODEVALUES "" ' "TYPE BASE END_GLYPH"
        ).statements
        self.assertEqual(
            (
                def_glyph.name,
                def_glyph.id,
                def_glyph.unicode,
                def_glyph.type,
                def_glyph.components,
            ),
            ("i.locl", 269, None, "BASE", None),
        )

    def test_def_glyph_base_2_components(self):
        [def_glyph] = self.parse(
            'DEF_GLYPH "glyphBase" ID 320 TYPE BASE COMPONENTS 2 END_GLYPH'
        ).statements
        self.assertEqual(
            (
                def_glyph.name,
                def_glyph.id,
                def_glyph.unicode,
                def_glyph.type,
                def_glyph.components,
            ),
            ("glyphBase", 320, None, "BASE", 2),
        )

    def test_def_glyph_ligature_2_components(self):
        [def_glyph] = self.parse(
            'DEF_GLYPH "f_f" ID 320 TYPE LIGATURE COMPONENTS 2 END_GLYPH'
        ).statements
        self.assertEqual(
            (
                def_glyph.name,
                def_glyph.id,
                def_glyph.unicode,
                def_glyph.type,
                def_glyph.components,
            ),
            ("f_f", 320, None, "LIGATURE", 2),
        )

    def test_def_glyph_mark(self):
        [def_glyph] = self.parse(
            'DEF_GLYPH "brevecomb" ID 320 TYPE MARK END_GLYPH'
        ).statements
        self.assertEqual(
            (
                def_glyph.name,
                def_glyph.id,
                def_glyph.unicode,
                def_glyph.type,
                def_glyph.components,
            ),
            ("brevecomb", 320, None, "MARK", None),
        )

    def test_def_glyph_component(self):
        [def_glyph] = self.parse(
            'DEF_GLYPH "f.f_f" ID 320 TYPE COMPONENT END_GLYPH'
        ).statements
        self.assertEqual(
            (
                def_glyph.name,
                def_glyph.id,
                def_glyph.unicode,
                def_glyph.type,
                def_glyph.components,
            ),
            ("f.f_f", 320, None, "COMPONENT", None),
        )

    def test_def_glyph_no_type(self):
        [def_glyph] = self.parse('DEF_GLYPH "glyph20" ID 20 END_GLYPH').statements
        self.assertEqual(
            (
                def_glyph.name,
                def_glyph.id,
                def_glyph.unicode,
                def_glyph.type,
                def_glyph.components,
            ),
            ("glyph20", 20, None, None, None),
        )

    def test_def_glyph_case_sensitive(self):
        def_glyphs = self.parse(
            'DEF_GLYPH "A" ID 3 UNICODE 65 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "a" ID 4 UNICODE 97 TYPE BASE END_GLYPH'
        ).statements
        self.assertEqual(
            (
                def_glyphs[0].name,
                def_glyphs[0].id,
                def_glyphs[0].unicode,
                def_glyphs[0].type,
                def_glyphs[0].components,
            ),
            ("A", 3, [0x41], "BASE", None),
        )
        self.assertEqual(
            (
                def_glyphs[1].name,
                def_glyphs[1].id,
                def_glyphs[1].unicode,
                def_glyphs[1].type,
                def_glyphs[1].components,
            ),
            ("a", 4, [0x61], "BASE", None),
        )

    def test_def_group_glyphs(self):
        [def_group] = self.parse(
            'DEF_GROUP "aaccented"\n'
            ' ENUM GLYPH "aacute" GLYPH "abreve" GLYPH "acircumflex" '
            'GLYPH "adieresis" GLYPH "ae" GLYPH "agrave" GLYPH "amacron" '
            'GLYPH "aogonek" GLYPH "aring" GLYPH "atilde" END_ENUM\n'
            "END_GROUP"
        ).statements
        self.assertEqual(
            (def_group.name, def_group.enum.glyphSet()),
            (
                "aaccented",
                (
                    "aacute",
                    "abreve",
                    "acircumflex",
                    "adieresis",
                    "ae",
                    "agrave",
                    "amacron",
                    "aogonek",
                    "aring",
                    "atilde",
                ),
            ),
        )

    def test_def_group_groups(self):
        [group1, group2, test_group] = self.parse(
            'DEF_GROUP "Group1"\n'
            ' ENUM GLYPH "a" GLYPH "b" GLYPH "c" GLYPH "d" END_ENUM\n'
            "END_GROUP\n"
            'DEF_GROUP "Group2"\n'
            ' ENUM GLYPH "e" GLYPH "f" GLYPH "g" GLYPH "h" END_ENUM\n'
            "END_GROUP\n"
            'DEF_GROUP "TestGroup"\n'
            ' ENUM GROUP "Group1" GROUP "Group2" END_ENUM\n'
            "END_GROUP"
        ).statements
        groups = [g.group for g in test_group.enum.enum]
        self.assertEqual((test_group.name, groups), ("TestGroup", ["Group1", "Group2"]))

    def test_def_group_groups_not_yet_defined(self):
        [group1, test_group1, test_group2, test_group3, group2] = self.parse(
            'DEF_GROUP "Group1"\n'
            ' ENUM GLYPH "a" GLYPH "b" GLYPH "c" GLYPH "d" END_ENUM\n'
            "END_GROUP\n"
            'DEF_GROUP "TestGroup1"\n'
            ' ENUM GROUP "Group1" GROUP "Group2" END_ENUM\n'
            "END_GROUP\n"
            'DEF_GROUP "TestGroup2"\n'
            ' ENUM GROUP "Group2" END_ENUM\n'
            "END_GROUP\n"
            'DEF_GROUP "TestGroup3"\n'
            ' ENUM GROUP "Group2" GROUP "Group1" END_ENUM\n'
            "END_GROUP\n"
            'DEF_GROUP "Group2"\n'
            ' ENUM GLYPH "e" GLYPH "f" GLYPH "g" GLYPH "h" END_ENUM\n'
            "END_GROUP"
        ).statements
        groups = [g.group for g in test_group1.enum.enum]
        self.assertEqual(
            (test_group1.name, groups), ("TestGroup1", ["Group1", "Group2"])
        )
        groups = [g.group for g in test_group2.enum.enum]
        self.assertEqual((test_group2.name, groups), ("TestGroup2", ["Group2"]))
        groups = [g.group for g in test_group3.enum.enum]
        self.assertEqual(
            (test_group3.name, groups), ("TestGroup3", ["Group2", "Group1"])
        )

    # def test_def_group_groups_undefined(self):
    #     with self.assertRaisesRegex(
    #             VoltLibError,
    #             r'Group "Group2" is used but undefined.'):
    #         [group1, test_group, group2] = self.parse(
    #             'DEF_GROUP "Group1"\n'
    #             'ENUM GLYPH "a" GLYPH "b" GLYPH "c" GLYPH "d" END_ENUM\n'
    #             'END_GROUP\n'
    #             'DEF_GROUP "TestGroup"\n'
    #             'ENUM GROUP "Group1" GROUP "Group2" END_ENUM\n'
    #             'END_GROUP\n'
    #         ).statements

    def test_def_group_glyphs_and_group(self):
        [def_group1, def_group2] = self.parse(
            'DEF_GROUP "aaccented"\n'
            ' ENUM GLYPH "aacute" GLYPH "abreve" GLYPH "acircumflex" '
            'GLYPH "adieresis" GLYPH "ae" GLYPH "agrave" GLYPH "amacron" '
            'GLYPH "aogonek" GLYPH "aring" GLYPH "atilde" END_ENUM\n'
            "END_GROUP\n"
            'DEF_GROUP "KERN_lc_a_2ND"\n'
            ' ENUM GLYPH "a" GROUP "aaccented" END_ENUM\n'
            "END_GROUP"
        ).statements
        items = def_group2.enum.enum
        self.assertEqual(
            (def_group2.name, items[0].glyphSet(), items[1].group),
            ("KERN_lc_a_2ND", ("a",), "aaccented"),
        )

    def test_def_group_range(self):
        def_group = self.parse(
            'DEF_GLYPH "a" ID 163 UNICODE 97 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "agrave" ID 194 UNICODE 224 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "aacute" ID 195 UNICODE 225 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "acircumflex" ID 196 UNICODE 226 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "atilde" ID 197 UNICODE 227 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "c" ID 165 UNICODE 99 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "ccaron" ID 209 UNICODE 269 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "ccedilla" ID 210 UNICODE 231 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "cdotaccent" ID 210 UNICODE 267 TYPE BASE END_GLYPH\n'
            'DEF_GROUP "KERN_lc_a_2ND"\n'
            ' ENUM RANGE "a" TO "atilde" GLYPH "b" RANGE "c" TO "cdotaccent" '
            "END_ENUM\n"
            "END_GROUP"
        ).statements[-1]
        self.assertEqual(
            (def_group.name, def_group.enum.glyphSet()),
            (
                "KERN_lc_a_2ND",
                (
                    "a",
                    "agrave",
                    "aacute",
                    "acircumflex",
                    "atilde",
                    "b",
                    "c",
                    "ccaron",
                    "ccedilla",
                    "cdotaccent",
                ),
            ),
        )

    def test_group_duplicate(self):
        self.assertRaisesRegex(
            VoltLibError,
            'Glyph group "dupe" already defined, ' "group names are case insensitive",
            self.parse,
            'DEF_GROUP "dupe"\n'
            'ENUM GLYPH "a" GLYPH "b" END_ENUM\n'
            "END_GROUP\n"
            'DEF_GROUP "dupe"\n'
            'ENUM GLYPH "x" END_ENUM\n'
            "END_GROUP",
        )

    def test_group_duplicate_case_insensitive(self):
        self.assertRaisesRegex(
            VoltLibError,
            'Glyph group "Dupe" already defined, ' "group names are case insensitive",
            self.parse,
            'DEF_GROUP "dupe"\n'
            'ENUM GLYPH "a" GLYPH "b" END_ENUM\n'
            "END_GROUP\n"
            'DEF_GROUP "Dupe"\n'
            'ENUM GLYPH "x" END_ENUM\n'
            "END_GROUP",
        )

    def test_script_without_langsys(self):
        [script] = self.parse(
            'DEF_SCRIPT NAME "Latin" TAG "latn"\n\n' "END_SCRIPT"
        ).statements
        self.assertEqual((script.name, script.tag, script.langs), ("Latin", "latn", []))

    def test_langsys_normal(self):
        [def_script] = self.parse(
            'DEF_SCRIPT NAME "Latin" TAG "latn"\n\n'
            'DEF_LANGSYS NAME "Romanian" TAG "ROM "\n\n'
            "END_LANGSYS\n"
            'DEF_LANGSYS NAME "Moldavian" TAG "MOL "\n\n'
            "END_LANGSYS\n"
            "END_SCRIPT"
        ).statements
        self.assertEqual((def_script.name, def_script.tag), ("Latin", "latn"))
        def_lang = def_script.langs[0]
        self.assertEqual((def_lang.name, def_lang.tag), ("Romanian", "ROM "))
        def_lang = def_script.langs[1]
        self.assertEqual((def_lang.name, def_lang.tag), ("Moldavian", "MOL "))

    def test_langsys_no_script_name(self):
        [langsys] = self.parse(
            'DEF_SCRIPT TAG "latn"\n\n'
            'DEF_LANGSYS NAME "Default" TAG "dflt"\n\n'
            "END_LANGSYS\n"
            "END_SCRIPT"
        ).statements
        self.assertEqual((langsys.name, langsys.tag), (None, "latn"))
        lang = langsys.langs[0]
        self.assertEqual((lang.name, lang.tag), ("Default", "dflt"))

    def test_langsys_no_script_tag_fails(self):
        with self.assertRaisesRegex(VoltLibError, r'.*Expected "TAG"'):
            [langsys] = self.parse(
                'DEF_SCRIPT NAME "Latin"\n\n'
                'DEF_LANGSYS NAME "Default" TAG "dflt"\n\n'
                "END_LANGSYS\n"
                "END_SCRIPT"
            ).statements

    def test_langsys_duplicate_script(self):
        with self.assertRaisesRegex(
            VoltLibError,
            'Script "DFLT" already defined, ' "script tags are case insensitive",
        ):
            [langsys1, langsys2] = self.parse(
                'DEF_SCRIPT NAME "Default" TAG "DFLT"\n\n'
                'DEF_LANGSYS NAME "Default" TAG "dflt"\n\n'
                "END_LANGSYS\n"
                "END_SCRIPT\n"
                'DEF_SCRIPT TAG "DFLT"\n\n'
                'DEF_LANGSYS NAME "Default" TAG "dflt"\n\n'
                "END_LANGSYS\n"
                "END_SCRIPT"
            ).statements

    def test_langsys_duplicate_lang(self):
        with self.assertRaisesRegex(
            VoltLibError,
            'Language "dflt" already defined in script "DFLT", '
            "language tags are case insensitive",
        ):
            [langsys] = self.parse(
                'DEF_SCRIPT NAME "Default" TAG "DFLT"\n'
                'DEF_LANGSYS NAME "Default" TAG "dflt"\n'
                "END_LANGSYS\n"
                'DEF_LANGSYS NAME "Default" TAG "dflt"\n'
                "END_LANGSYS\n"
                "END_SCRIPT"
            ).statements

    def test_langsys_lang_in_separate_scripts(self):
        [langsys1, langsys2] = self.parse(
            'DEF_SCRIPT NAME "Default" TAG "DFLT"\n\n'
            'DEF_LANGSYS NAME "Default" TAG "dflt"\n\n'
            "END_LANGSYS\n"
            'DEF_LANGSYS NAME "Default" TAG "ROM "\n\n'
            "END_LANGSYS\n"
            "END_SCRIPT\n"
            'DEF_SCRIPT NAME "Latin" TAG "latn"\n\n'
            'DEF_LANGSYS NAME "Default" TAG "dflt"\n\n'
            "END_LANGSYS\n"
            'DEF_LANGSYS NAME "Default" TAG "ROM "\n\n'
            "END_LANGSYS\n"
            "END_SCRIPT"
        ).statements
        self.assertEqual(
            (langsys1.langs[0].tag, langsys1.langs[1].tag), ("dflt", "ROM ")
        )
        self.assertEqual(
            (langsys2.langs[0].tag, langsys2.langs[1].tag), ("dflt", "ROM ")
        )

    def test_langsys_no_lang_name(self):
        [langsys] = self.parse(
            'DEF_SCRIPT NAME "Latin" TAG "latn"\n\n'
            'DEF_LANGSYS TAG "dflt"\n\n'
            "END_LANGSYS\n"
            "END_SCRIPT"
        ).statements
        self.assertEqual((langsys.name, langsys.tag), ("Latin", "latn"))
        lang = langsys.langs[0]
        self.assertEqual((lang.name, lang.tag), (None, "dflt"))

    def test_langsys_no_langsys_tag_fails(self):
        with self.assertRaisesRegex(VoltLibError, r'.*Expected "TAG"'):
            [langsys] = self.parse(
                'DEF_SCRIPT NAME "Latin" TAG "latn"\n\n'
                'DEF_LANGSYS NAME "Default"\n\n'
                "END_LANGSYS\n"
                "END_SCRIPT"
            ).statements

    def test_feature(self):
        [def_script] = self.parse(
            'DEF_SCRIPT NAME "Latin" TAG "latn"\n\n'
            'DEF_LANGSYS NAME "Romanian" TAG "ROM "\n\n'
            'DEF_FEATURE NAME "Fractions" TAG "frac"\n'
            ' LOOKUP "fraclookup"\n'
            "END_FEATURE\n"
            "END_LANGSYS\n"
            "END_SCRIPT"
        ).statements
        def_feature = def_script.langs[0].features[0]
        self.assertEqual(
            (def_feature.name, def_feature.tag, def_feature.lookups),
            ("Fractions", "frac", ["fraclookup"]),
        )
        [def_script] = self.parse(
            'DEF_SCRIPT NAME "Latin" TAG "latn"\n\n'
            'DEF_LANGSYS NAME "Romanian" TAG "ROM "\n\n'
            'DEF_FEATURE NAME "Kerning" TAG "kern"\n'
            ' LOOKUP "kern1" LOOKUP "kern2"\n'
            "END_FEATURE\n"
            "END_LANGSYS\n"
            "END_SCRIPT"
        ).statements
        def_feature = def_script.langs[0].features[0]
        self.assertEqual(
            (def_feature.name, def_feature.tag, def_feature.lookups),
            ("Kerning", "kern", ["kern1", "kern2"]),
        )

    def test_lookup_duplicate(self):
        with self.assertRaisesRegex(
            VoltLibError,
            'Lookup "dupe" already defined, ' "lookup names are case insensitive",
        ):
            [lookup1, lookup2] = self.parse(
                'DEF_LOOKUP "dupe"\n'
                "AS_SUBSTITUTION\n"
                'SUB GLYPH "a"\n'
                'WITH GLYPH "a.alt"\n'
                "END_SUB\n"
                "END_SUBSTITUTION\n"
                'DEF_LOOKUP "dupe"\n'
                "AS_SUBSTITUTION\n"
                'SUB GLYPH "b"\n'
                'WITH GLYPH "b.alt"\n'
                "END_SUB\n"
                "END_SUBSTITUTION\n"
            ).statements

    def test_lookup_duplicate_insensitive_case(self):
        with self.assertRaisesRegex(
            VoltLibError,
            'Lookup "Dupe" already defined, ' "lookup names are case insensitive",
        ):
            [lookup1, lookup2] = self.parse(
                'DEF_LOOKUP "dupe"\n'
                "AS_SUBSTITUTION\n"
                'SUB GLYPH "a"\n'
                'WITH GLYPH "a.alt"\n'
                "END_SUB\n"
                "END_SUBSTITUTION\n"
                'DEF_LOOKUP "Dupe"\n'
                "AS_SUBSTITUTION\n"
                'SUB GLYPH "b"\n'
                'WITH GLYPH "b.alt"\n'
                "END_SUB\n"
                "END_SUBSTITUTION\n"
            ).statements

    def test_lookup_name_starts_with_letter(self):
        with self.assertRaisesRegex(
            VoltLibError, r'Lookup name "\\lookupname" must start with a letter'
        ):
            [lookup] = self.parse(
                'DEF_LOOKUP "\\lookupname"\n'
                "AS_SUBSTITUTION\n"
                'SUB GLYPH "a"\n'
                'WITH GLYPH "a.alt"\n'
                "END_SUB\n"
                "END_SUBSTITUTION\n"
            ).statements

    def test_lookup_comments(self):
        [lookup] = self.parse(
            'DEF_LOOKUP "test" PROCESS_BASE PROCESS_MARKS ALL DIRECTION LTR\n'
            'COMMENTS "Hello\\nWorld"\n'
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "a"\n'
            'WITH GLYPH "b"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        ).statements
        self.assertEqual(lookup.name, "test")
        self.assertEqual(lookup.comments, "Hello\nWorld")

    def test_substitution_empty(self):
        with self.assertRaisesRegex(VoltLibError, r"Expected SUB"):
            [lookup] = self.parse(
                'DEF_LOOKUP "empty_substitution" PROCESS_BASE PROCESS_MARKS '
                "ALL DIRECTION LTR\n"
                "IN_CONTEXT\n"
                "END_CONTEXT\n"
                "AS_SUBSTITUTION\n"
                "END_SUBSTITUTION"
            ).statements

    def test_substitution_invalid_many_to_many(self):
        with self.assertRaisesRegex(VoltLibError, r"Invalid substitution type"):
            [lookup] = self.parse(
                'DEF_LOOKUP "invalid_substitution" PROCESS_BASE PROCESS_MARKS '
                "ALL DIRECTION LTR\n"
                "IN_CONTEXT\n"
                "END_CONTEXT\n"
                "AS_SUBSTITUTION\n"
                'SUB GLYPH "f" GLYPH "i"\n'
                'WITH GLYPH "f.alt" GLYPH "i.alt"\n'
                "END_SUB\n"
                "END_SUBSTITUTION"
            ).statements

    def test_substitution_invalid_reverse_chaining_single(self):
        [lookup] = self.parse(
            'DEF_LOOKUP "invalid_substitution" PROCESS_BASE PROCESS_MARKS '
            "ALL DIRECTION LTR REVERSAL\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "f" GLYPH "i"\n'
            'WITH GLYPH "f_i"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        ).statements
        self.assertIsInstance(lookup.sub, ast.SubstitutionLigatureDefinition)

    def test_substitution_invalid_mixed(self):
        with self.assertRaisesRegex(VoltLibError, r"Invalid substitution type"):
            [lookup] = self.parse(
                'DEF_LOOKUP "invalid_substitution" PROCESS_BASE PROCESS_MARKS '
                "ALL DIRECTION LTR\n"
                "IN_CONTEXT\n"
                "END_CONTEXT\n"
                "AS_SUBSTITUTION\n"
                'SUB GLYPH "fi"\n'
                'WITH GLYPH "f" GLYPH "i"\n'
                "END_SUB\n"
                'SUB GLYPH "f" GLYPH "l"\n'
                'WITH GLYPH "f_l"\n'
                "END_SUB\n"
                "END_SUBSTITUTION"
            ).statements

    def test_substitution_single(self):
        [lookup] = self.parse(
            'DEF_LOOKUP "smcp" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "a"\n'
            'WITH GLYPH "a.sc"\n'
            "END_SUB\n"
            'SUB GLYPH "b"\n'
            'WITH GLYPH "b.sc"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        ).statements
        self.assertEqual(lookup.name, "smcp")
        self.assertSubEqual(lookup.sub, [["a"], ["b"]], [["a.sc"], ["b.sc"]])

    def test_substitution_single_in_context(self):
        [group, lookup] = self.parse(
            'DEF_GROUP "Denominators"\n'
            ' ENUM GLYPH "one.dnom" GLYPH "two.dnom" END_ENUM\n'
            "END_GROUP\n"
            'DEF_LOOKUP "fracdnom" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            ' LEFT ENUM GROUP "Denominators" GLYPH "fraction" END_ENUM\n'
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "one"\n'
            'WITH GLYPH "one.dnom"\n'
            "END_SUB\n"
            'SUB GLYPH "two"\n'
            'WITH GLYPH "two.dnom"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        ).statements
        context = lookup.context[0]

        self.assertEqual(lookup.name, "fracdnom")
        self.assertEqual(context.ex_or_in, "IN_CONTEXT")
        self.assertEqual(len(context.left), 1)
        self.assertEqual(len(context.left[0]), 1)
        self.assertEqual(len(context.left[0][0].enum), 2)
        self.assertEqual(context.left[0][0].enum[0].group, "Denominators")
        self.assertEqual(context.left[0][0].enum[1].glyph, "fraction")
        self.assertEqual(context.right, [])
        self.assertSubEqual(
            lookup.sub, [["one"], ["two"]], [["one.dnom"], ["two.dnom"]]
        )

    def test_substitution_single_in_contexts(self):
        [group, lookup] = self.parse(
            'DEF_GROUP "Hebrew"\n'
            ' ENUM GLYPH "uni05D0" GLYPH "uni05D1" END_ENUM\n'
            "END_GROUP\n"
            'DEF_LOOKUP "HebrewCurrency" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            ' RIGHT GROUP "Hebrew"\n'
            ' RIGHT GLYPH "one.Hebr"\n'
            "END_CONTEXT\n"
            "IN_CONTEXT\n"
            ' LEFT GROUP "Hebrew"\n'
            ' LEFT GLYPH "one.Hebr"\n'
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "dollar"\n'
            'WITH GLYPH "dollar.Hebr"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        ).statements
        context1 = lookup.context[0]
        context2 = lookup.context[1]

        self.assertEqual(lookup.name, "HebrewCurrency")

        self.assertEqual(context1.ex_or_in, "IN_CONTEXT")
        self.assertEqual(context1.left, [])
        self.assertEqual(len(context1.right), 2)
        self.assertEqual(len(context1.right[0]), 1)
        self.assertEqual(len(context1.right[1]), 1)
        self.assertEqual(context1.right[0][0].group, "Hebrew")
        self.assertEqual(context1.right[1][0].glyph, "one.Hebr")

        self.assertEqual(context2.ex_or_in, "IN_CONTEXT")
        self.assertEqual(len(context2.left), 2)
        self.assertEqual(len(context2.left[0]), 1)
        self.assertEqual(len(context2.left[1]), 1)
        self.assertEqual(context2.left[0][0].group, "Hebrew")
        self.assertEqual(context2.left[1][0].glyph, "one.Hebr")
        self.assertEqual(context2.right, [])

    def test_substitution_skip_base(self):
        [group, lookup] = self.parse(
            'DEF_GROUP "SomeMarks"\n'
            ' ENUM GLYPH "marka" GLYPH "markb" END_ENUM\n'
            "END_GROUP\n"
            'DEF_LOOKUP "SomeSub" SKIP_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "A"\n'
            'WITH GLYPH "A.c2sc"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        ).statements
        self.assertEqual((lookup.name, lookup.process_base), ("SomeSub", False))

    def test_substitution_process_base(self):
        [group, lookup] = self.parse(
            'DEF_GROUP "SomeMarks"\n'
            ' ENUM GLYPH "marka" GLYPH "markb" END_ENUM\n'
            "END_GROUP\n"
            'DEF_LOOKUP "SomeSub" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "A"\n'
            'WITH GLYPH "A.c2sc"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        ).statements
        self.assertEqual((lookup.name, lookup.process_base), ("SomeSub", True))

    def test_substitution_process_marks(self):
        [group, lookup] = self.parse(
            'DEF_GROUP "SomeMarks"\n'
            ' ENUM GLYPH "marka" GLYPH "markb" END_ENUM\n'
            "END_GROUP\n"
            'DEF_LOOKUP "SomeSub" PROCESS_BASE PROCESS_MARKS "SomeMarks"\n'
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "A"\n'
            'WITH GLYPH "A.c2sc"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        ).statements
        self.assertEqual((lookup.name, lookup.process_marks), ("SomeSub", "SomeMarks"))

    def test_substitution_process_marks_all(self):
        [lookup] = self.parse(
            'DEF_LOOKUP "SomeSub" PROCESS_BASE PROCESS_MARKS ALL\n'
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "A"\n'
            'WITH GLYPH "A.c2sc"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        ).statements
        self.assertEqual((lookup.name, lookup.process_marks), ("SomeSub", True))

    def test_substitution_process_marks_none(self):
        [lookup] = self.parse_(
            'DEF_LOOKUP "SomeSub" PROCESS_BASE PROCESS_MARKS "NONE"\n'
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "A"\n'
            'WITH GLYPH "A.c2sc"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        ).statements
        self.assertEqual((lookup.name, lookup.process_marks), ("SomeSub", False))

    def test_substitution_process_marks_bad(self):
        with self.assertRaisesRegex(
            VoltLibError, "Expected ALL, NONE, MARK_GLYPH_SET or an ID"
        ):
            self.parse(
                'DEF_GROUP "SomeMarks" ENUM GLYPH "marka" GLYPH "markb" '
                "END_ENUM END_GROUP\n"
                'DEF_LOOKUP "SomeSub" PROCESS_BASE PROCESS_MARKS SomeMarks '
                "AS_SUBSTITUTION\n"
                'SUB GLYPH "A" WITH GLYPH "A.c2sc"\n'
                "END_SUB\n"
                "END_SUBSTITUTION"
            )

    def test_substitution_skip_marks(self):
        [group, lookup] = self.parse(
            'DEF_GROUP "SomeMarks"\n'
            ' ENUM GLYPH "marka" GLYPH "markb" END_ENUM\n'
            "END_GROUP\n"
            'DEF_LOOKUP "SomeSub" PROCESS_BASE SKIP_MARKS DIRECTION LTR\n'
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "A"\n'
            'WITH GLYPH "A.c2sc"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        ).statements
        self.assertEqual((lookup.name, lookup.process_marks), ("SomeSub", False))

    def test_substitution_mark_attachment(self):
        [group, lookup] = self.parse(
            'DEF_GROUP "SomeMarks"\n'
            ' ENUM GLYPH "acutecmb" GLYPH "gravecmb" END_ENUM\n'
            "END_GROUP\n"
            'DEF_LOOKUP "SomeSub" PROCESS_BASE '
            'PROCESS_MARKS "SomeMarks" DIRECTION RTL\n'
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "A"\n'
            'WITH GLYPH "A.c2sc"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        ).statements
        self.assertEqual((lookup.name, lookup.process_marks), ("SomeSub", "SomeMarks"))

    def test_substitution_mark_glyph_set(self):
        [group, lookup] = self.parse(
            'DEF_GROUP "SomeMarks"\n'
            ' ENUM GLYPH "acutecmb" GLYPH "gravecmb" END_ENUM\n'
            "END_GROUP\n"
            'DEF_LOOKUP "SomeSub" PROCESS_BASE '
            'PROCESS_MARKS MARK_GLYPH_SET "SomeMarks" DIRECTION RTL\n'
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "A"\n'
            'WITH GLYPH "A.c2sc"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        ).statements
        self.assertEqual((lookup.name, lookup.mark_glyph_set), ("SomeSub", "SomeMarks"))

    def test_substitution_process_all_marks(self):
        [group, lookup] = self.parse(
            'DEF_GROUP "SomeMarks"\n'
            ' ENUM GLYPH "acutecmb" GLYPH "gravecmb" END_ENUM\n'
            "END_GROUP\n"
            'DEF_LOOKUP "SomeSub" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION RTL\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "A"\n'
            'WITH GLYPH "A.c2sc"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        ).statements
        self.assertEqual((lookup.name, lookup.process_marks), ("SomeSub", True))

    def test_substitution_no_reversal(self):
        # TODO: check right context with no reversal
        [lookup] = self.parse(
            'DEF_LOOKUP "Lookup" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            ' RIGHT ENUM GLYPH "a" GLYPH "b" END_ENUM\n'
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "a"\n'
            'WITH GLYPH "a.alt"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        ).statements
        self.assertEqual((lookup.name, lookup.reversal), ("Lookup", None))

    def test_substitution_reversal(self):
        lookup = self.parse(
            'DEF_GROUP "DFLT_Num_standardFigures"\n'
            ' ENUM GLYPH "zero" GLYPH "one" GLYPH "two" END_ENUM\n'
            "END_GROUP\n"
            'DEF_GROUP "DFLT_Num_numerators"\n'
            ' ENUM GLYPH "zero.numr" GLYPH "one.numr" GLYPH "two.numr" END_ENUM\n'
            "END_GROUP\n"
            'DEF_LOOKUP "RevLookup" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR REVERSAL\n"
            "IN_CONTEXT\n"
            ' RIGHT ENUM GLYPH "a" GLYPH "b" END_ENUM\n'
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GROUP "DFLT_Num_standardFigures"\n'
            'WITH GROUP "DFLT_Num_numerators"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        ).statements[-1]
        self.assertEqual((lookup.name, lookup.reversal), ("RevLookup", True))

    def test_substitution_single_to_multiple(self):
        [lookup] = self.parse(
            'DEF_LOOKUP "ccmp" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "aacute"\n'
            'WITH GLYPH "a" GLYPH "acutecomb"\n'
            "END_SUB\n"
            'SUB GLYPH "agrave"\n'
            'WITH GLYPH "a" GLYPH "gravecomb"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        ).statements
        self.assertEqual(lookup.name, "ccmp")
        self.assertSubEqual(
            lookup.sub,
            [["aacute"], ["agrave"]],
            [["a", "acutecomb"], ["a", "gravecomb"]],
        )

    def test_substitution_multiple_to_single(self):
        [lookup] = self.parse(
            'DEF_LOOKUP "liga" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "f" GLYPH "i"\n'
            'WITH GLYPH "f_i"\n'
            "END_SUB\n"
            'SUB GLYPH "f" GLYPH "t"\n'
            'WITH GLYPH "f_t"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        ).statements
        self.assertEqual(lookup.name, "liga")
        self.assertSubEqual(lookup.sub, [["f", "i"], ["f", "t"]], [["f_i"], ["f_t"]])

    def test_substitution_reverse_chaining_single(self):
        lookup = self.parse(
            'DEF_GLYPH "zero" ID 163 UNICODE 48 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "one" ID 194 UNICODE 49 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "two" ID 195 UNICODE 50 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "three" ID 196 UNICODE 51 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "four" ID 197 UNICODE 52 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "fize" ID 165 UNICODE 53 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "six" ID 209 UNICODE 54 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "seven" ID 210 UNICODE 55 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "eight" ID 211 UNICODE 56 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "nine" ID 212 UNICODE 57 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "zero.numr" ID 213 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "one.numr" ID 214 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "two.numr" ID 215 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "three.numr" ID 216 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "four.numr" ID 217 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "fize.numr" ID 218 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "six.numr" ID 219 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "seven.numr" ID 220 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "eight.numr" ID 221 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "nine.numr" ID 222 TYPE BASE END_GLYPH\n'
            'DEF_LOOKUP "numr" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR REVERSAL\n"
            "IN_CONTEXT\n"
            " RIGHT ENUM "
            'GLYPH "fraction" '
            'RANGE "zero.numr" TO "nine.numr" '
            "END_ENUM\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB RANGE "zero" TO "nine"\n'
            'WITH RANGE "zero.numr" TO "nine.numr"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        ).statements[-1]

        mapping = lookup.sub.mapping
        glyphs = [[(r.start, r.end) for r in v] for v in mapping.keys()]
        replacement = [[(r.start, r.end) for r in v] for v in mapping.values()]

        self.assertEqual(lookup.name, "numr")
        self.assertEqual(glyphs, [[("zero", "nine")]])
        self.assertEqual(replacement, [[("zero.numr", "nine.numr")]])

        self.assertEqual(len(lookup.context[0].right), 1)
        self.assertEqual(len(lookup.context[0].right[0]), 1)
        enum = lookup.context[0].right[0][0]
        self.assertEqual(len(enum.enum), 2)
        self.assertEqual(enum.enum[0].glyph, "fraction")
        self.assertEqual(
            (enum.enum[1].start, enum.enum[1].end), ("zero.numr", "nine.numr")
        )

    # GPOS
    #  ATTACH_CURSIVE
    #  ATTACH
    #  ADJUST_PAIR
    #  ADJUST_SINGLE
    def test_position_empty(self):
        with self.assertRaisesRegex(
            VoltLibError, "Expected ATTACH, ATTACH_CURSIVE, ADJUST_PAIR, ADJUST_SINGLE"
        ):
            [lookup] = self.parse(
                'DEF_LOOKUP "empty_position" PROCESS_BASE PROCESS_MARKS ALL '
                "DIRECTION LTR\n"
                "EXCEPT_CONTEXT\n"
                ' LEFT GLYPH "glyph"\n'
                "END_CONTEXT\n"
                "AS_POSITION\n"
                "END_POSITION"
            ).statements

    def test_position_attach(self):
        [lookup, anchor1, anchor2, anchor3, anchor4] = self.parse(
            'DEF_LOOKUP "anchor_top" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION RTL\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_POSITION\n"
            'ATTACH GLYPH "a" GLYPH "e"\n'
            'TO GLYPH "acutecomb" AT ANCHOR "top" '
            'GLYPH "gravecomb" AT ANCHOR "top"\n'
            "END_ATTACH\n"
            "END_POSITION\n"
            'DEF_ANCHOR "MARK_top" ON 120 GLYPH acutecomb COMPONENT 1 '
            "AT  POS DX 0 DY 450 END_POS END_ANCHOR\n"
            'DEF_ANCHOR "MARK_top" ON 121 GLYPH gravecomb COMPONENT 1 '
            "AT  POS DX 0 DY 450 END_POS END_ANCHOR\n"
            'DEF_ANCHOR "top" ON 31 GLYPH a COMPONENT 1 '
            "AT  POS DX 210 DY 450 END_POS END_ANCHOR\n"
            'DEF_ANCHOR "top" ON 35 GLYPH e COMPONENT 1 '
            "AT  POS DX 215 DY 450 END_POS END_ANCHOR"
        ).statements
        pos = lookup.pos
        coverage = [g.glyph for g in pos.coverage]
        coverage_to = [[[g.glyph for g in e], a] for (e, a) in pos.coverage_to]
        self.assertEqual(
            (lookup.name, coverage, coverage_to),
            (
                "anchor_top",
                ["a", "e"],
                [[["acutecomb"], "top"], [["gravecomb"], "top"]],
            ),
        )
        self.assertEqual(
            (
                anchor1.name,
                anchor1.gid,
                anchor1.glyph_name,
                anchor1.component,
                anchor1.locked,
                anchor1.pos,
            ),
            ("MARK_top", 120, "acutecomb", 1, False, (None, 0, 450, {}, {}, {})),
        )
        self.assertEqual(
            (
                anchor2.name,
                anchor2.gid,
                anchor2.glyph_name,
                anchor2.component,
                anchor2.locked,
                anchor2.pos,
            ),
            ("MARK_top", 121, "gravecomb", 1, False, (None, 0, 450, {}, {}, {})),
        )
        self.assertEqual(
            (
                anchor3.name,
                anchor3.gid,
                anchor3.glyph_name,
                anchor3.component,
                anchor3.locked,
                anchor3.pos,
            ),
            ("top", 31, "a", 1, False, (None, 210, 450, {}, {}, {})),
        )
        self.assertEqual(
            (
                anchor4.name,
                anchor4.gid,
                anchor4.glyph_name,
                anchor4.component,
                anchor4.locked,
                anchor4.pos,
            ),
            ("top", 35, "e", 1, False, (None, 215, 450, {}, {}, {})),
        )

    def test_position_attach_cursive(self):
        [lookup] = self.parse(
            'DEF_LOOKUP "SomeLookup" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION RTL\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_POSITION\n"
            'ATTACH_CURSIVE\nEXIT  GLYPH "a" GLYPH "b"\nENTER  GLYPH "c"\n'
            "END_ATTACH\n"
            "END_POSITION"
        ).statements
        exit = [[g.glyph for g in v] for v in lookup.pos.coverages_exit]
        enter = [[g.glyph for g in v] for v in lookup.pos.coverages_enter]
        self.assertEqual(
            (lookup.name, exit, enter), ("SomeLookup", [["a", "b"]], [["c"]])
        )

    def test_position_adjust_pair(self):
        [lookup] = self.parse(
            'DEF_LOOKUP "kern1" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION RTL\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_POSITION\n"
            "ADJUST_PAIR\n"
            ' FIRST  GLYPH "A"\n'
            ' SECOND  GLYPH "V"\n'
            " 1 2 BY POS ADV -30 END_POS POS END_POS\n"
            " 2 1 BY POS ADV -30 END_POS POS END_POS\n\n"
            "END_ADJUST\n"
            "END_POSITION"
        ).statements
        coverages_1 = [[g.glyph for g in v] for v in lookup.pos.coverages_1]
        coverages_2 = [[g.glyph for g in v] for v in lookup.pos.coverages_2]
        self.assertEqual(
            (lookup.name, coverages_1, coverages_2, lookup.pos.adjust_pair),
            (
                "kern1",
                [["A"]],
                [["V"]],
                {
                    (1, 2): (
                        (-30, None, None, {}, {}, {}),
                        (None, None, None, {}, {}, {}),
                    ),
                    (2, 1): (
                        (-30, None, None, {}, {}, {}),
                        (None, None, None, {}, {}, {}),
                    ),
                },
            ),
        )

    def test_position_adjust_single(self):
        [lookup] = self.parse(
            'DEF_LOOKUP "TestLookup" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            # ' LEFT GLYPH "leftGlyph"\n'
            # ' RIGHT GLYPH "rightGlyph"\n'
            "END_CONTEXT\n"
            "AS_POSITION\n"
            "ADJUST_SINGLE"
            ' GLYPH "glyph1" BY POS ADV 0 DX 123 END_POS'
            ' GLYPH "glyph2" BY POS ADV 0 DX 456 END_POS\n'
            "END_ADJUST\n"
            "END_POSITION"
        ).statements
        pos = lookup.pos
        adjust = [[[g.glyph for g in a], b] for (a, b) in pos.adjust_single]
        self.assertEqual(
            (lookup.name, adjust),
            (
                "TestLookup",
                [
                    [["glyph1"], (0, 123, None, {}, {}, {})],
                    [["glyph2"], (0, 456, None, {}, {}, {})],
                ],
            ),
        )

    def test_def_anchor(self):
        [anchor1, anchor2, anchor3] = self.parse(
            'DEF_ANCHOR "top" ON 120 GLYPH a '
            "COMPONENT 1 AT  POS DX 250 DY 450 END_POS END_ANCHOR\n"
            'DEF_ANCHOR "MARK_top" ON 120 GLYPH acutecomb '
            "COMPONENT 1 AT  POS DX 0 DY 450 END_POS END_ANCHOR\n"
            'DEF_ANCHOR "bottom" ON 120 GLYPH a '
            "COMPONENT 1 AT  POS DX 250 DY 0 END_POS END_ANCHOR"
        ).statements
        self.assertEqual(
            (
                anchor1.name,
                anchor1.gid,
                anchor1.glyph_name,
                anchor1.component,
                anchor1.locked,
                anchor1.pos,
            ),
            ("top", 120, "a", 1, False, (None, 250, 450, {}, {}, {})),
        )
        self.assertEqual(
            (
                anchor2.name,
                anchor2.gid,
                anchor2.glyph_name,
                anchor2.component,
                anchor2.locked,
                anchor2.pos,
            ),
            ("MARK_top", 120, "acutecomb", 1, False, (None, 0, 450, {}, {}, {})),
        )
        self.assertEqual(
            (
                anchor3.name,
                anchor3.gid,
                anchor3.glyph_name,
                anchor3.component,
                anchor3.locked,
                anchor3.pos,
            ),
            ("bottom", 120, "a", 1, False, (None, 250, 0, {}, {}, {})),
        )

    def test_def_anchor_multi_component(self):
        [anchor1, anchor2] = self.parse(
            'DEF_ANCHOR "top" ON 120 GLYPH a '
            "COMPONENT 1 AT  POS DX 250 DY 450 END_POS END_ANCHOR\n"
            'DEF_ANCHOR "top" ON 120 GLYPH a '
            "COMPONENT 2 AT  POS DX 250 DY 450 END_POS END_ANCHOR"
        ).statements
        self.assertEqual(
            (anchor1.name, anchor1.gid, anchor1.glyph_name, anchor1.component),
            ("top", 120, "a", 1),
        )
        self.assertEqual(
            (anchor2.name, anchor2.gid, anchor2.glyph_name, anchor2.component),
            ("top", 120, "a", 2),
        )

    def test_def_anchor_duplicate(self):
        self.assertRaisesRegex(
            VoltLibError,
            'Anchor "dupe" already defined, ' "anchor names are case insensitive",
            self.parse,
            'DEF_ANCHOR "dupe" ON 120 GLYPH a '
            "COMPONENT 1 AT  POS DX 250 DY 450 END_POS END_ANCHOR\n"
            'DEF_ANCHOR "dupe" ON 120 GLYPH a '
            "COMPONENT 1 AT  POS DX 250 DY 450 END_POS END_ANCHOR",
        )

    def test_def_anchor_locked(self):
        [anchor] = self.parse(
            'DEF_ANCHOR "top" ON 120 GLYPH a '
            "COMPONENT 1 LOCKED AT  POS DX 250 DY 450 END_POS END_ANCHOR"
        ).statements
        self.assertEqual(
            (
                anchor.name,
                anchor.gid,
                anchor.glyph_name,
                anchor.component,
                anchor.locked,
                anchor.pos,
            ),
            ("top", 120, "a", 1, True, (None, 250, 450, {}, {}, {})),
        )

    def test_anchor_adjust_device(self):
        [anchor] = self.parse(
            'DEF_ANCHOR "MARK_top" ON 123 GLYPH diacglyph '
            "COMPONENT 1 AT  POS DX 0 DY 456 ADJUST_BY 12 AT 34 "
            "ADJUST_BY 56 AT 78 END_POS END_ANCHOR"
        ).statements
        self.assertEqual(
            (anchor.name, anchor.pos),
            ("MARK_top", (None, 0, 456, {}, {}, {34: 12, 78: 56})),
        )

    def test_ppem(self):
        [grid_ppem, pres_ppem, ppos_ppem] = self.parse(
            "GRID_PPEM 20\n" "PRESENTATION_PPEM 72\n" "PPOSITIONING_PPEM 144"
        ).statements
        self.assertEqual(
            (
                (grid_ppem.name, grid_ppem.value),
                (pres_ppem.name, pres_ppem.value),
                (ppos_ppem.name, ppos_ppem.value),
            ),
            (("GRID_PPEM", 20), ("PRESENTATION_PPEM", 72), ("PPOSITIONING_PPEM", 144)),
        )

    def test_compiler_flags(self):
        [setting1, setting2] = self.parse(
            "COMPILER_USEEXTENSIONLOOKUPS\n" "COMPILER_USEPAIRPOSFORMAT2"
        ).statements
        self.assertEqual(
            ((setting1.name, setting1.value), (setting2.name, setting2.value)),
            (
                ("COMPILER_USEEXTENSIONLOOKUPS", True),
                ("COMPILER_USEPAIRPOSFORMAT2", True),
            ),
        )

    def test_cmap(self):
        [cmap_format1, cmap_format2, cmap_format3] = self.parse(
            "CMAP_FORMAT 0 3 4\n" "CMAP_FORMAT 1 0 6\n" "CMAP_FORMAT 3 1 4"
        ).statements
        self.assertEqual(
            (
                (cmap_format1.name, cmap_format1.value),
                (cmap_format2.name, cmap_format2.value),
                (cmap_format3.name, cmap_format3.value),
            ),
            (
                ("CMAP_FORMAT", (0, 3, 4)),
                ("CMAP_FORMAT", (1, 0, 6)),
                ("CMAP_FORMAT", (3, 1, 4)),
            ),
        )

    def test_do_not_touch_cmap(self):
        [option1, option2, option3, option4] = self.parse(
            "DO_NOT_TOUCH_CMAP\n"
            "CMAP_FORMAT 0 3 4\n"
            "CMAP_FORMAT 1 0 6\n"
            "CMAP_FORMAT 3 1 4"
        ).statements
        self.assertEqual(
            (
                (option1.name, option1.value),
                (option2.name, option2.value),
                (option3.name, option3.value),
                (option4.name, option4.value),
            ),
            (
                ("DO_NOT_TOUCH_CMAP", True),
                ("CMAP_FORMAT", (0, 3, 4)),
                ("CMAP_FORMAT", (1, 0, 6)),
                ("CMAP_FORMAT", (3, 1, 4)),
            ),
        )

    def test_stop_at_end(self):
        doc = self.parse_('DEF_GLYPH ".notdef" ID 0 TYPE BASE END_GLYPH END\0\0\0\0')
        [def_glyph] = doc.statements
        self.assertEqual(
            (
                def_glyph.name,
                def_glyph.id,
                def_glyph.unicode,
                def_glyph.type,
                def_glyph.components,
            ),
            (".notdef", 0, None, "BASE", None),
        )
        self.assertEqual(
            str(doc), '\nDEF_GLYPH ".notdef" ID 0 TYPE BASE END_GLYPH END\n'
        )

    def parse_(self, text):
        return Parser(StringIO(text)).parse()

    def parse(self, text):
        doc = self.parse_(text)
        self.assertEqual("\n".join(str(s) for s in doc.statements), text)
        return Parser(StringIO(text)).parse()


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



import io
import os
import re
import random
import tempfile
from fontTools.feaLib.builder import addOpenTypeFeaturesFromString
from fontTools.ttLib import (
    TTFont,
    TTLibError,
    newTable,
    registerCustomTableClass,
    unregisterCustomTableClass,
)
from fontTools.ttLib.standardGlyphOrder import standardGlyphOrder
from fontTools.ttLib.tables.DefaultTable import DefaultTable
from fontTools.ttLib.tables._c_m_a_p import CmapSubtable
import pytest


DATA_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), "data")


class CustomTableClass(DefaultTable):
    def decompile(self, data, ttFont):
        self.numbers = list(data)

    def compile(self, ttFont):
        return bytes(self.numbers)

    # not testing XML read/write


table_C_U_S_T_ = CustomTableClass  # alias for testing


TABLETAG = "CUST"


def normalize_TTX(string):
    string = re.sub(' ttLibVersion=".*"', "", string)
    string = re.sub('checkSumAdjustment value=".*"', "", string)
    string = re.sub('modified value=".*"', "", string)
    return string


def test_registerCustomTableClass():
    font = TTFont()
    font[TABLETAG] = newTable(TABLETAG)
    font[TABLETAG].data = b"\x00\x01\xff"
    f = io.BytesIO()
    font.save(f)
    f.seek(0)
    assert font[TABLETAG].data == b"\x00\x01\xff"
    registerCustomTableClass(TABLETAG, "ttFont_test", "CustomTableClass")
    try:
        font = TTFont(f)
        assert font[TABLETAG].numbers == [0, 1, 255]
        assert font[TABLETAG].compile(font) == b"\x00\x01\xff"
    finally:
        unregisterCustomTableClass(TABLETAG)


def test_registerCustomTableClassStandardName():
    registerCustomTableClass(TABLETAG, "ttFont_test")
    try:
        font = TTFont()
        font[TABLETAG] = newTable(TABLETAG)
        font[TABLETAG].numbers = [4, 5, 6]
        assert font[TABLETAG].compile(font) == b"\x04\x05\x06"
    finally:
        unregisterCustomTableClass(TABLETAG)


ttxTTF = r"""<?xml version="1.0" encoding="UTF-8"?>
<ttFont sfntVersion="\x00\x01\x00\x00" ttLibVersion="4.9.0">
  <hmtx>
    <mtx name=".notdef" width="300" lsb="0"/>
  </hmtx>
</ttFont>
"""


ttxOTF = """<?xml version="1.0" encoding="UTF-8"?>
<ttFont sfntVersion="OTTO" ttLibVersion="4.9.0">
  <hmtx>
    <mtx name=".notdef" width="300" lsb="0"/>
  </hmtx>
</ttFont>
"""


def test_sfntVersionFromTTX():
    # https://github.com/fonttools/fonttools/issues/2370
    font = TTFont()
    assert font.sfntVersion == "\x00\x01\x00\x00"
    ttx = io.StringIO(ttxOTF)
    # Font is "empty", TTX file will determine sfntVersion
    font.importXML(ttx)
    assert font.sfntVersion == "OTTO"
    ttx = io.StringIO(ttxTTF)
    # Font is not "empty", sfntVersion in TTX file will be ignored
    font.importXML(ttx)
    assert font.sfntVersion == "OTTO"


def test_virtualGlyphId():
    otfpath = os.path.join(DATA_DIR, "TestVGID-Regular.otf")
    ttxpath = os.path.join(DATA_DIR, "TestVGID-Regular.ttx")

    otf = TTFont(otfpath)

    ttx = TTFont()
    ttx.importXML(ttxpath)

    with open(ttxpath, encoding="utf-8") as fp:
        xml = normalize_TTX(fp.read()).splitlines()

    for font in (otf, ttx):
        GSUB = font["GSUB"].table
        assert GSUB.LookupList.LookupCount == 37
        lookup = GSUB.LookupList.Lookup[32]
        assert lookup.LookupType == 8
        subtable = lookup.SubTable[0]
        assert subtable.LookAheadGlyphCount == 1
        lookahead = subtable.LookAheadCoverage[0]
        assert len(lookahead.glyphs) == 46
        assert "glyph00453" in lookahead.glyphs

        out = io.StringIO()
        font.saveXML(out)
        outxml = normalize_TTX(out.getvalue()).splitlines()
        assert xml == outxml


def test_setGlyphOrder_also_updates_glyf_glyphOrder():
    # https://github.com/fonttools/fonttools/issues/2060#issuecomment-1063932428
    font = TTFont()
    font.importXML(os.path.join(DATA_DIR, "TestTTF-Regular.ttx"))
    current_order = font.getGlyphOrder()

    assert current_order == font["glyf"].glyphOrder

    new_order = list(current_order)
    while new_order == current_order:
        random.shuffle(new_order)

    font.setGlyphOrder(new_order)

    assert font.getGlyphOrder() == new_order
    assert font["glyf"].glyphOrder == new_order


def test_getGlyphOrder_not_true_post_format_1(caplog):
    # https://github.com/fonttools/fonttools/issues/2736
    caplog.set_level("WARNING")
    font = TTFont(os.path.join(DATA_DIR, "bogus_post_format_1.ttf"))
    hmtx = font["hmtx"]
    assert len(hmtx.metrics) > len(standardGlyphOrder)
    log_rec = caplog.records[-1]
    assert log_rec.levelname == "WARNING"
    assert "Not enough names found in the 'post' table" in log_rec.message


@pytest.mark.parametrize("lazy", [None, True, False])
def test_ensureDecompiled(lazy):
    # test that no matter the lazy value, ensureDecompiled decompiles all tables
    font = TTFont()
    font.importXML(os.path.join(DATA_DIR, "TestTTF-Regular.ttx"))
    # test font has no OTL so we add some, as an example of otData-driven tables
    addOpenTypeFeaturesFromString(
        font,
        """
        feature calt {
            sub period' period' period' space by ellipsis;
        } calt;

        feature dist {
            pos period period -30;
        } dist;
        """,
    )
    # also add an additional cmap subtable that will be lazily-loaded
    cm = CmapSubtable.newSubtable(14)
    cm.platformID = 0
    cm.platEncID = 5
    cm.language = 0
    cm.cmap = {}
    cm.uvsDict = {0xFE00: [(0x002E, None)]}
    font["cmap"].tables.append(cm)

    # save and reload, potentially lazily
    buf = io.BytesIO()
    font.save(buf)
    buf.seek(0)
    font = TTFont(buf, lazy=lazy)

    # check no table is loaded until/unless requested, no matter the laziness
    for tag in font.keys():
        assert not font.isLoaded(tag)

    if lazy is not False:
        # additional cmap doesn't get decompiled automatically unless lazy=False;
        # can't use hasattr or else cmap's maginc __getattr__ kicks in...
        cm = next(st for st in font["cmap"].tables if st.__dict__["format"] == 14)
        assert cm.data is not None
        assert "uvsDict" not in cm.__dict__
        # glyf glyphs are not expanded unless lazy=False
        assert font["glyf"].glyphs["period"].data is not None
        assert not hasattr(font["glyf"].glyphs["period"], "coordinates")

    if lazy is True:
        # OTL tables hold a 'reader' to lazily load when lazy=True
        assert "reader" in font["GSUB"].table.LookupList.__dict__
        assert "reader" in font["GPOS"].table.LookupList.__dict__

    font.ensureDecompiled()

    # all tables are decompiled now
    for tag in font.keys():
        assert font.isLoaded(tag)
    # including the additional cmap
    cm = next(st for st in font["cmap"].tables if st.__dict__["format"] == 14)
    assert cm.data is None
    assert "uvsDict" in cm.__dict__
    # expanded glyf glyphs lost the 'data' attribute
    assert not hasattr(font["glyf"].glyphs["period"], "data")
    assert hasattr(font["glyf"].glyphs["period"], "coordinates")
    # and OTL tables have read their 'reader'
    assert "reader" not in font["GSUB"].table.LookupList.__dict__
    assert "Lookup" in font["GSUB"].table.LookupList.__dict__
    assert "reader" not in font["GPOS"].table.LookupList.__dict__
    assert "Lookup" in font["GPOS"].table.LookupList.__dict__


@pytest.fixture
def testFont_fvar_avar():
    ttxpath = os.path.join(DATA_DIR, "TestTTF_normalizeLocation.ttx")
    ttf = TTFont()
    ttf.importXML(ttxpath)
    return ttf


@pytest.mark.parametrize(
    "userLocation, expectedNormalizedLocation",
    [
        ({}, {"wght": 0.0}),
        ({"wght": 100}, {"wght": -1.0}),
        ({"wght": 250}, {"wght": -0.75}),
        ({"wght": 400}, {"wght": 0.0}),
        ({"wght": 550}, {"wght": 0.75}),
        ({"wght": 625}, {"wght": 0.875}),
        ({"wght": 700}, {"wght": 1.0}),
    ],
)
def test_font_normalizeLocation(
    testFont_fvar_avar, userLocation, expectedNormalizedLocation
):
    normalizedLocation = testFont_fvar_avar.normalizeLocation(userLocation)
    assert expectedNormalizedLocation == normalizedLocation


def test_font_normalizeLocation_no_VF():
    ttf = TTFont()
    with pytest.raises(TTLibError, match="Not a variable font"):
        ttf.normalizeLocation({})


def test_getGlyphID():
    font = TTFont()
    font.importXML(os.path.join(DATA_DIR, "TestTTF-Regular.ttx"))

    assert font.getGlyphID("space") == 3
    assert font.getGlyphID("glyph12345") == 12345  # virtual glyph
    with pytest.raises(KeyError):
        font.getGlyphID("non_existent")
    with pytest.raises(KeyError):
        font.getGlyphID("glyph_prefix_but_invalid_id")


def test_spooled_tempfile_may_not_have_attribute_seekable():
    # SpooledTemporaryFile only got a seekable attribute on Python 3.11
    # https://github.com/fonttools/fonttools/issues/3052
    font = TTFont()
    font.importXML(os.path.join(DATA_DIR, "TestTTF-Regular.ttx"))
    tmp = tempfile.SpooledTemporaryFile()
    font.save(tmp)
    # this should not fail
    _ = TTFont(tmp)


def test_unseekable_file_lazy_loading_fails():
    class NonSeekableFile:
        def __init__(self):
            self.file = io.BytesIO()

        def read(self, size):
            return self.file.read(size)

        def seekable(self):
            return False

    f = NonSeekableFile()
    with pytest.raises(TTLibError, match="Input file must be seekable when lazy=True"):
        TTFont(f, lazy=True)


def test_unsupported_seek_operation_lazy_loading_fails():
    class UnsupportedSeekFile:
        def __init__(self):
            self.file = io.BytesIO()

        def read(self, size):
            return self.file.read(size)

        def seek(self, offset):
            raise io.UnsupportedOperation("Unsupported seek operation")

    f = UnsupportedSeekFile()
    with pytest.raises(TTLibError, match="Input file must be seekable when lazy=True"):
        TTFont(f, lazy=True)


@pytest.mark.parametrize(
    "file_name",
    [
        "duplicate_glyph_name.ttf",
        "duplicate_glyph_name.otf",
    ],
)
def test_duplicate_glyph_names(file_name):
    font_path = os.path.join(DATA_DIR, file_name)
    font = TTFont(font_path)

    assert font.getGlyphOrder() == [".notdef", "space", "A", "A.2", "A.1"]

    if "CFF " not in font:
        post = font["post"]
        assert post.mapping == {"A.2": "A"}


def test_saveXML_default_dumps_all_tables():
    """tables=None (default) should dump all tables (regression test for #4025)."""
    font = TTFont()
    font.importXML(os.path.join(DATA_DIR, "TestTTF-Regular.ttx"))
    out = io.StringIO()
    font.saveXML(out)
    xml = out.getvalue()
    assert "<head>" in xml
    assert "<hhea>" in xml


def test_saveXML_specific_tables():
    """Passing tables=['head'] should dump only head."""
    font = TTFont()
    font.importXML(os.path.join(DATA_DIR, "TestTTF-Regular.ttx"))
    out = io.StringIO()
    font.saveXML(out, tables=["head"])
    xml = out.getvalue()
    assert "<head>" in xml
    assert "<hhea>" not in xml


def test_saveXML_skipTables():
    """skipTables should exclude specified tables when tables is None."""
    font = TTFont()
    font.importXML(os.path.join(DATA_DIR, "TestTTF-Regular.ttx"))
    out = io.StringIO()
    font.saveXML(out, skipTables=["head"])
    xml = out.getvalue()
    assert "<head>" not in xml
    assert "<hhea>" in xml



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from fontTools.pens.qu2cuPen import Qu2CuPen
from fontTools.pens.recordingPen import RecordingPen
from textwrap import dedent
import pytest

try:
    from .utils import CUBIC_GLYPHS, QUAD_GLYPHS
    from .utils import DummyGlyph
    from .utils import DummyPen
except ImportError as e:
    pytest.skip(str(e), allow_module_level=True)

MAX_ERR = 1.0


class _TestPenMixin(object):
    """Collection of tests that are shared by both the SegmentPen and the
    PointPen test cases, plus some helper methods.
    Note: We currently don't have a PointPen.
    """

    maxDiff = None

    def diff(self, expected, actual):
        import difflib

        expected = str(self.Glyph(expected)).splitlines(True)
        actual = str(self.Glyph(actual)).splitlines(True)
        diff = difflib.unified_diff(
            expected, actual, fromfile="expected", tofile="actual"
        )
        return "".join(diff)

    def convert_glyph(self, glyph, **kwargs):
        # draw source glyph onto a new glyph using a Cu2Qu pen and return it
        converted = self.Glyph()
        pen = getattr(converted, self.pen_getter_name)()
        cubicpen = self.Qu2CuPen(pen, MAX_ERR, all_cubic=True, **kwargs)
        getattr(glyph, self.draw_method_name)(cubicpen)
        return converted

    def expect_glyph(self, source, expected):
        converted = self.convert_glyph(source)
        self.assertNotEqual(converted, source)
        if not converted.approx(expected):
            print(self.diff(expected, converted))
            self.fail("converted glyph is different from expected")

    def test_convert_simple_glyph(self):
        self.expect_glyph(QUAD_GLYPHS["a"], CUBIC_GLYPHS["a"])
        self.expect_glyph(QUAD_GLYPHS["A"], CUBIC_GLYPHS["A"])

    def test_convert_composite_glyph(self):
        source = CUBIC_GLYPHS["Aacute"]
        converted = self.convert_glyph(source)
        # components don't change after quadratic conversion
        self.assertEqual(converted, source)

    def test_reverse_direction(self):
        for name in ("a", "A", "Eacute"):
            source = QUAD_GLYPHS[name]
            normal_glyph = self.convert_glyph(source)
            reversed_glyph = self.convert_glyph(source, reverse_direction=True)

            # the number of commands is the same, just their order is iverted
            self.assertTrue(len(normal_glyph.outline), len(reversed_glyph.outline))
            self.assertNotEqual(normal_glyph, reversed_glyph)

    def test_stats(self):
        stats = {}
        for name in QUAD_GLYPHS.keys():
            source = QUAD_GLYPHS[name]
            self.convert_glyph(source, stats=stats)

        self.assertTrue(stats)
        self.assertTrue("2" in stats)
        self.assertEqual(type(stats["2"]), int)

    def test_addComponent(self):
        pen = self.Pen()
        cubicpen = self.Qu2CuPen(pen, MAX_ERR)
        cubicpen.addComponent("a", (1, 2, 3, 4, 5.0, 6.0))

        # components are passed through without changes
        self.assertEqual(
            str(pen).splitlines(),
            [
                "pen.addComponent('a', (1, 2, 3, 4, 5.0, 6.0))",
            ],
        )


class TestQu2CuPen(unittest.TestCase, _TestPenMixin):
    def __init__(self, *args, **kwargs):
        super(TestQu2CuPen, self).__init__(*args, **kwargs)
        self.Glyph = DummyGlyph
        self.Pen = DummyPen
        self.Qu2CuPen = Qu2CuPen
        self.pen_getter_name = "getPen"
        self.draw_method_name = "draw"

    def test_qCurveTo_1_point(self):
        pen = DummyPen()
        cubicpen = Qu2CuPen(pen, MAX_ERR)
        cubicpen.moveTo((0, 0))
        cubicpen.qCurveTo((1, 1))
        cubicpen.closePath()

        self.assertEqual(
            str(pen).splitlines(),
            [
                "pen.moveTo((0, 0))",
                "pen.qCurveTo((1, 1))",
                "pen.closePath()",
            ],
        )

    def test_qCurveTo_2_points(self):
        pen = DummyPen()
        cubicpen = Qu2CuPen(pen, MAX_ERR)
        cubicpen.moveTo((0, 0))
        cubicpen.qCurveTo((1, 1), (2, 2))
        cubicpen.closePath()

        self.assertEqual(
            str(pen).splitlines(),
            [
                "pen.moveTo((0, 0))",
                "pen.qCurveTo((1, 1), (2, 2))",
                "pen.closePath()",
            ],
        )

    def test_qCurveTo_3_points_no_conversion(self):
        pen = DummyPen()
        cubicpen = Qu2CuPen(pen, MAX_ERR)
        cubicpen.moveTo((0, 0))
        cubicpen.qCurveTo((0, 3), (1, 3), (1, 0))
        cubicpen.closePath()

        self.assertEqual(
            str(pen).splitlines(),
            [
                "pen.moveTo((0, 0))",
                "pen.qCurveTo((0, 3), (1, 3), (1, 0))",
                "pen.closePath()",
            ],
        )

    def test_qCurveTo_no_oncurve_points(self):
        pen = DummyPen()
        cubicpen = Qu2CuPen(pen, MAX_ERR)
        cubicpen.qCurveTo((0, 0), (1, 0), (1, 1), (0, 1), None)
        cubicpen.closePath()

        self.assertEqual(
            str(pen).splitlines(),
            ["pen.qCurveTo((0, 0), (1, 0), (1, 1), (0, 1), None)", "pen.closePath()"],
        )

    def test_curveTo_1_point(self):
        pen = DummyPen()
        cubicpen = Qu2CuPen(pen, MAX_ERR)
        cubicpen.moveTo((0, 0))
        cubicpen.curveTo((1, 1))
        cubicpen.closePath()

        self.assertEqual(
            str(pen).splitlines(),
            [
                "pen.moveTo((0, 0))",
                "pen.curveTo((1, 1))",
                "pen.closePath()",
            ],
        )

    def test_curveTo_2_points(self):
        pen = DummyPen()
        cubicpen = Qu2CuPen(pen, MAX_ERR)
        cubicpen.moveTo((0, 0))
        cubicpen.curveTo((1, 1), (2, 2))
        cubicpen.closePath()

        self.assertEqual(
            str(pen).splitlines(),
            [
                "pen.moveTo((0, 0))",
                "pen.curveTo((1, 1), (2, 2))",
                "pen.closePath()",
            ],
        )

    def test_curveTo_3_points(self):
        pen = DummyPen()
        cubicpen = Qu2CuPen(pen, MAX_ERR)
        cubicpen.moveTo((0, 0))
        cubicpen.curveTo((1, 1), (2, 2), (3, 3))
        cubicpen.closePath()

        self.assertEqual(
            str(pen).splitlines(),
            [
                "pen.moveTo((0, 0))",
                "pen.curveTo((1, 1), (2, 2), (3, 3))",
                "pen.closePath()",
            ],
        )

    def test_all_cubic(self):
        inPen = RecordingPen()
        inPen.value = [
            ("moveTo", ((1204, 347),)),
            ("qCurveTo", ((1255, 347), (1323, 433), (1323, 467))),
            ("qCurveTo", ((1323, 478), (1310, 492), (1302, 492))),
            ("qCurveTo", ((1295, 492), (1289, 484))),
            ("lineTo", ((1272, 461),)),
            ("qCurveTo", ((1256, 439), (1221, 416), (1200, 416))),
            ("qCurveTo", ((1181, 416), (1141, 440), (1141, 462))),
            ("qCurveTo", ((1141, 484), (1190, 565), (1190, 594))),
            ("qCurveTo", ((1190, 607), (1181, 634), (1168, 634))),
            ("qCurveTo", ((1149, 634), (1146, 583), (1081, 496), (1081, 463))),
            ("qCurveTo", ((1081, 417), (1164, 347), (1204, 347))),
            ("closePath", ()),
        ]

        outPen = RecordingPen()
        q2cPen = Qu2CuPen(outPen, 1.0, all_cubic=True)
        inPen.replay(q2cPen)

        print(outPen.value)

        assert not any(typ == "qCurveTo" for typ, _ in outPen.value)


if __name__ == "__main__":
    unittest.main()
